class BreakInnerLoop(Exception):
    """
    raised to break all inner loop, either of Idler or of RunWith or WithoutStateMachine
    """
    pass

class BreakAllLoops(Exception):
    """
    raised to exit the event loop itself
    """
    pass

class StateMachineError(Exception):
    """
    raise to show errors while calling actions or writing properties in wrong state
    """
    pass 

class DatabaseError(Exception):
    """
    raise to show database related errors
    """



__all__ = ['BreakInnerLoop', 'BreakAllLoops', 'StateMachineError']


from collections import OrderedDict
import sys
import inspect
import typing
from functools import reduce, partial


def classlist(class_ : typing.Any) -> typing.Tuple[type]:
    """
    Return a list of the class hierarchy above (and including) the given class.

    Same as `inspect.getmro(class_)[::-1]`
    """
    return inspect.getmro(class_)[::-1]
    

def get_dot_resolved_attr(obj : typing.Any, attr : str, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)
    return reduce(_getattr, [obj] + attr.split('.'))


def iscoroutinefunction(function : typing.Callable) -> bool:
    """
    Whether the function is an asynchronous coroutine function.
    """
    import asyncio
    try:
        return (
            inspect.isasyncgenfunction(function) or
            asyncio.iscoroutinefunction(function)
        )
    except AttributeError:
        return False
    

def get_method_owner(method : typing.Callable) -> typing.Any:
    """
    Gets the instance that owns the supplied method
    """
    if not inspect.ismethod(method):
        return None
    if isinstance(method, partial):
        method = method.func
    return method.__self__ if sys.version_info.major >= 3 else method.im_self


def is_ordered_dict(d):
    """
    Predicate checking for ordered dictionaries. OrderedDict is always
    ordered, and vanilla Python dictionaries are ordered for Python 3.6+
    """
    py3_ordered_dicts = (sys.version_info.major == 3) and (sys.version_info.minor >= 6)
    vanilla_odicts = (sys.version_info.major > 3) or py3_ordered_dicts
    return isinstance(d, OrderedDict)or (vanilla_odicts and isinstance(d, dict))


def get_all_slots(class_):
    """
    Return a list of slot names for slots defined in `class_` and its
    superclasses.
    """
    # A subclass's __slots__ attribute does not contain slots defined
    # in its superclass (the superclass' __slots__ end up as
    # attributes of the subclass).
    all_slots = []
    parent_param_classes = [c for c in classlist(class_)[1::]]
    for c in parent_param_classes:
        if hasattr(c,'__slots__'):
            all_slots+=c.__slots__
    return all_slots


def get_occupied_slots(instance):
    """
    Return a list of slots for which values have been set.

    (While a slot might be defined, if a value for that slot hasn't
    been set, then it's an AttributeError to request the slot's
    value.)
    """
    return [slot for slot in get_all_slots(type(instance))
            if hasattr(instance, slot)]




__all__ = ['classlist', 'get_dot_resolved_attr', 'iscoroutinefunction', 'get_method_owner', 'get_all_slots', 
        'get_occupied_slots']


import logging 
import inspect
import os
import ssl
import typing
import warnings
import zmq
import zmq.asyncio

from ..param.parameterized import Parameterized, ParameterizedMetaclass, edit_constant as edit_constant_parameters
from .constants import (JSON, LOGLEVEL, ZMQ_PROTOCOLS, HTTP_METHODS, JSONSerializable)
from .database import ThingDB, ThingInformation
from .serializers import _get_serializer_from_user_given_options, BaseSerializer, JSONSerializer
from .schema_validators import BaseSchemaValidator, JsonSchemaValidator
from .exceptions import BreakInnerLoop
from .action import action
from .dataklasses import HTTPResource, ZMQResource, build_our_temp_TD, get_organised_resources
from .utils import get_default_logger, getattr_without_descriptor_read
from .property import Property, ClassProperties
from .properties import String, ClassSelector, Selector, TypedKeyMappingsConstrainedDict
from .zmq_message_brokers import RPCServer, ServerTypes, EventPublisher
from .state_machine import StateMachine
from .events import Event




class ThingMeta(ParameterizedMetaclass):
    """
    Metaclass for Thing, implements a ``__post_init__()`` call and instantiation of a container for properties' descriptor 
    objects. During instantiation of ``Thing``, first serializers, loggers and database connection are created, after which
    the user ``__init__`` is called. In ``__post_init__()``, that runs after user's ``__init__()``, the exposed resources 
    are segregated while accounting for any ``Event`` objects or instance specific properties created during init. Properties 
    are also loaded from database at this time. One can overload ``__post_init__()`` for any operations that rely on properties
    values loaded from database.
    """
    
    @classmethod
    def __prepare__(cls, name, bases):
        return TypedKeyMappingsConstrainedDict({},
            type_mapping = dict(
                state_machine = (StateMachine, type(None)),
                instance_name = String, 
                log_level = Selector,
                logger = ClassSelector,
                logfile = String,
                db_config_file = String,
                object_info = Property, # it should not be set by the user
            ),
            allow_unspecified_keys = True
        )

    def __new__(cls, __name, __bases, __dict : TypedKeyMappingsConstrainedDict):
        return super().__new__(cls, __name, __bases, __dict._inner)
    
    def __call__(mcls, *args, **kwargs):
        instance = super().__call__(*args, **kwargs)
        instance.__post_init__()
        return instance
    
    def _create_param_container(mcs, mcs_members : dict) -> None:
        """
        creates ``ClassProperties`` instead of ``param``'s own ``Parameters`` 
        as the default container for descriptors. All properties have definitions 
        copied from ``param``.
        """
        mcs._param_container = ClassProperties(mcs, mcs_members)

    @property
    def properties(mcs) -> ClassProperties:
        """
        returns ``ClassProperties`` instance instead of ``param``'s own 
        ``Parameters`` instance. See code of ``param``.
        """
        return mcs._param_container



class Thing(Parameterized, metaclass=ThingMeta):
    """
    Subclass from here to expose python objects on the network (with HTTP/TCP) or to other processes (ZeroMQ)
    """

    __server_type__ = ServerTypes.THING # not a server, this needs to be removed.
   
    # local properties
    instance_name = String(default=None, regex=r'[A-Za-z]+[A-Za-z_0-9\-\/]*', constant=True, remote=False,
                        doc="""Unique string identifier of the instance. This value is used for many operations,
                        for example - creating zmq socket address, tables in databases, and to identify the instance 
                        in the HTTP Server - (http(s)://{domain and sub domain}/{instance name}). 
                        If creating a big system, instance names are recommended to be unique.""") # type: str
    logger = ClassSelector(class_=logging.Logger, default=None, allow_None=True, remote=False, 
                        doc="""logging.Logger instance to print log messages. Default 
                            logger with a IO-stream handler and network accessible handler is created 
                            if none supplied.""") # type: logging.Logger
    zmq_serializer = ClassSelector(class_=(BaseSerializer, str), 
                        allow_None=True, default='json', remote=False,
                        doc="""Serializer used for exchanging messages with python RPC clients. Subclass the base serializer 
                        or one of the available serializers to implement your own serialization requirements; or, register 
                        type replacements. Default is JSON. Some serializers like MessagePack improve performance many times 
                        compared to JSON and can be useful for data intensive applications within python.""") # type: BaseSerializer
    http_serializer = ClassSelector(class_=(JSONSerializer, str), default=None, allow_None=True, remote=False,
                        doc="""Serializer used for exchanging messages with a HTTP clients,
                            subclass JSONSerializer to implement your own JSON serialization requirements; or, 
                            register type replacements. Other types of serializers are currently not allowed for HTTP clients.""") # type: JSONSerializer
    schema_validator = ClassSelector(class_=BaseSchemaValidator, default=JsonSchemaValidator, allow_None=True, 
                        remote=False, isinstance=False,
                        doc="""Validator for JSON schema. If not supplied, a default JSON schema validator is created.""") # type: BaseSchemaValidator
    
    # remote properties
    state = String(default=None, allow_None=True, URL_path='/state', readonly=True, observable=True, 
                fget=lambda self : self.state_machine.current_state if hasattr(self, 'state_machine') else None,  
                doc="current state machine's state if state machine present, None indicates absence of state machine.") #type: typing.Optional[str]
    httpserver_resources = Property(readonly=True, URL_path='/resources/http-server', 
                        doc="object's resources exposed to HTTP client (through ``hololinked.server.HTTPServer.HTTPServer``)", 
                        fget=lambda self: self._httpserver_resources ) # type: typing.Dict[str, HTTPResource]
    zmq_resources = Property(readonly=True, URL_path='/resources/zmq-object-proxy', 
                        doc="object's resources exposed to RPC client, similar to HTTP resources but differs in details.", 
                        fget=lambda self: self._zmq_resources) # type: typing.Dict[str, ZMQResource]
    gui_resources = Property(readonly=True, URL_path='/resources/portal-app', 
                        doc="""object's data read by hololinked-portal GUI client, similar to http_resources but differs 
                        in details.""",
                        fget=lambda self: build_our_temp_TD(self)) # type: typing.Dict[str, typing.Any]
    GUI = Property(default=None, allow_None=True, URL_path='/resources/web-gui', fget = lambda self : self._gui,
                        doc="GUI specified here will become visible at GUI tab of hololinked-portal dashboard tool")     
    object_info = Property(doc="contains information about this object like the class name, script location etc.",
                        URL_path='/object-info') # type: ThingInformation
    

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls)
        # defines some internal fixed attributes. attributes created by us that require no validation but 
        # cannot be modified are called _internal_fixed_attributes
        obj._internal_fixed_attributes = ['_internal_fixed_attributes', 'instance_resources',
                                        '_httpserver_resources', '_zmq_resources', '_owner', 'rpc_server', 'message_broker',
                                        '_event_publisher']        
        return obj


    def __init__(self, *, instance_name : str, logger : typing.Optional[logging.Logger] = None, 
                serializer : typing.Optional[JSONSerializer] = None, **kwargs) -> None:
        """
        Parameters
        ----------
        instance_name: str
            Unique string identifier of the instance. This value is used for many operations,
            for example - creating zmq socket address, tables in databases, and to identify the instance in the HTTP Server - 
            (http(s)://{domain and sub domain}/{instance name}). 
            If creating a big system, instance names are recommended to be unique.
        logger: logging.Logger, optional
            logging.Logger instance to print log messages. Default logger with a IO-stream handler and network 
            accessible handler is created if none supplied.
        serializer: JSONSerializer, optional
            custom JSON serializer. To use separate serializer for python RPC clients and cross-platform 
            HTTP clients, use keyword arguments zmq_serializer and http_serializer and leave this argument at None.
        **kwargs:
            zmq_serializer: BaseSerializer | str, optional 
                Serializer used for exchanging messages with python RPC clients. If string value is supplied, 
                supported are 'msgpack', 'pickle', 'serpent', 'json'. Subclass the base serializer 
                ``hololinked.server.serializer.BaseSerializer`` or one of the available serializers to implement your 
                own serialization requirements; or, register type replacements. Default is JSON. Some serializers like 
                MessagePack improve performance many times  compared to JSON and can be useful for data intensive 
                applications within python. The serializer supplied here must also be supplied to object proxy from 
                ``hololinked.client``. 
            http_serializer: JSONSerializer, optional
                serializer used for cross platform HTTP clients. 
            use_default_db: bool, Default False
                if True, default SQLite database is created where properties can be stored and loaded. There is no need to supply
                any database credentials. This value can also be set as a class attribute, see docs.
            logger_remote_access: bool, Default True
                if False, network accessible handler is not attached to the logger. This value can also be set as a 
                class attribute, see docs.
            schema_validator: BaseSchemaValidator, optional
                schema validator class for JSON schema validation, not supported by ZMQ clients. 
            db_config_file: str, optional
                if not using a default database, supply a JSON configuration file to create a connection. Check documentaion
                of ``hololinked.server.database``.  

        """
        if instance_name.startswith('/'):
            instance_name = instance_name[1:]
        # Type definitions
        self._owner : typing.Optional[Thing] = None 
        self._internal_fixed_attributes : typing.List[str]
        self._full_URL_path_prefix : str
        self.rpc_server  = None # type: typing.Optional[RPCServer]
        self.message_broker = None # type : typing.Optional[AsyncPollingZMQServer]
        self._event_publisher = None # type : typing.Optional[EventPublisher]
        self._gui = None # filler for a future feature
        # serializer
        if not isinstance(serializer, JSONSerializer) and serializer != 'json' and serializer is not None:
            raise TypeError("serializer key word argument must be JSONSerializer. If one wishes to use separate serializers " +
                            "for python clients and HTTP clients, use zmq_serializer and http_serializer keyword arguments.")
        zmq_serializer = serializer or kwargs.pop('zmq_serializer', 'json')
        http_serializer = serializer if isinstance(serializer, JSONSerializer) else kwargs.pop('http_serializer', 'json')
        zmq_serializer, http_serializer = _get_serializer_from_user_given_options(
                                                                    zmq_serializer=zmq_serializer,
                                                                    http_serializer=http_serializer
                                                                )
        super().__init__(instance_name=instance_name, logger=logger, 
                        zmq_serializer=zmq_serializer, http_serializer=http_serializer, **kwargs)

        self._prepare_logger(
                    log_level=kwargs.get('log_level', None), 
                    log_file=kwargs.get('log_file', None),
                    remote_access=kwargs.get('logger_remote_access', self.__class__.logger_remote_access if hasattr(
                                                                self.__class__, 'logger_remote_access') else False)
                )
        self._prepare_state_machine()  
        self._prepare_DB(kwargs.get('use_default_db', False), kwargs.get('db_config_file', None))   


    def __post_init__(self):
        self._prepare_resources()
        self.load_properties_from_DB()
        self.logger.info(f"initialialised Thing class {self.__class__.__name__} with instance name {self.instance_name}")


    def __setattr__(self, __name: str, __value: typing.Any) -> None:
        if  __name == '_internal_fixed_attributes' or __name in self._internal_fixed_attributes: 
            # order of 'or' operation for above 'if' matters
            if not hasattr(self, __name) or getattr(self, __name, None) is None:
                # allow setting of fixed attributes once
                super().__setattr__(__name, __value)
            else:
                raise AttributeError(f"Attempted to set {__name} more than once. " +
                                     "Cannot assign a value to this variable after creation.")
        else:
            super().__setattr__(__name, __value)


    def _prepare_resources(self):
        """
        this method analyses the members of the class which have '_remote_info' variable declared
        and extracts information necessary to make RPC functionality work.
        """
        # The following dict is to be given to the HTTP server
        self._zmq_resources, self._httpserver_resources, self.instance_resources = get_organised_resources(self)


    def _prepare_logger(self, log_level : int, log_file : str, remote_access : bool = False):
        from .logger import RemoteAccessHandler
        if self.logger is None:
            self.logger = get_default_logger(self.instance_name, 
                                    logging.INFO if not log_level else log_level, 
                                    None if not log_file else log_file)
        if remote_access:
            if not any(isinstance(handler, RemoteAccessHandler) for handler in self.logger.handlers):
                self._remote_access_loghandler = RemoteAccessHandler(instance_name='logger', 
                                                    maxlen=500, emit_interval=1, logger=self.logger) 
                                                    # thing has its own logger so we dont recreate one for
                                                    # remote access handler
                self.logger.addHandler(self._remote_access_loghandler)
        
        if not isinstance(self, logging.Logger):
            for handler in self.logger.handlers:
                # if remote access is True or not, if a default handler is found make a variable for it anyway
                if isinstance(handler, RemoteAccessHandler):
                    self._remote_access_loghandler = handler        


    def _prepare_state_machine(self):
        if hasattr(self, 'state_machine'):
            self.state_machine._prepare(self)
            self.logger.debug("setup state machine")

    
    def _prepare_DB(self, default_db : bool = False, config_file : str = None):
        if not default_db and not config_file: 
            self.object_info
            return 
        # 1. create engine 
        self.db_engine = ThingDB(instance=self, config_file=None if default_db else config_file, 
                                    serializer=self.zmq_serializer) # type: ThingDB 
        # 2. create an object metadata to be used by different types of clients
        object_info = self.db_engine.fetch_own_info()
        if object_info is not None:
            self._object_info = object_info
        # 3. enter properties to DB if not already present 
        if self.object_info.class_name != self.__class__.__name__:
            raise ValueError("Fetched instance name and class name from database not matching with the ", 
                "current Thing class/subclass. You might be reusing an instance name of another subclass ", 
                "and did not remove the old data from database. Please clean the database using database tools to ", 
                "start fresh.")


    @object_info.getter
    def _get_object_info(self):
        if not hasattr(self, '_object_info'):
            self._object_info = ThingInformation(
                    instance_name  = self.instance_name, 
                    class_name     = self.__class__.__name__,
                    script         = os.path.dirname(os.path.abspath(inspect.getfile(self.__class__))),
                    http_server    = "USER_MANAGED", 
                    kwargs         = "USER_MANAGED",  
                    eventloop_instance_name = "USER_MANAGED", 
                    level          = "USER_MANAGED", 
                    level_type     = "USER_MANAGED"
                )  
        return self._object_info
    
    @object_info.setter
    def _set_object_info(self, value):
        self._object_info = ThingInformation(**value)  

    
    @property
    def properties(self) -> ClassProperties:
        """container for the property descriptors of the object."""
        return self.parameters

    @action(URL_path='/properties', http_method=HTTP_METHODS.GET)
    def _get_properties(self, **kwargs) -> typing.Dict[str, typing.Any]:
        """
        """
        skip_props = ["httpserver_resources", "zmq_resources", "gui_resources", "GUI", "object_info"]
        for prop_name in skip_props:
            if prop_name in kwargs:
                raise RuntimeError("GUI, httpserver resources, RPC resources , object info etc. cannot be queried" + 
                                  " using multiple property fetch.")
        data = {}
        if len(kwargs) == 0:
            for name, prop in self.properties.descriptors.items():
                if name in skip_props or not isinstance(prop, Property):
                    continue
                if prop._remote_info is None:
                    continue
                data[name] = prop.__get__(self, type(self))
        elif 'names' in kwargs:
            names = kwargs.get('names')
            if not isinstance(names, (list, tuple, str)):
                raise TypeError(f"Specify properties to be fetched as a list, tuple or comma separated names. Givent type {type(names)}")
            if isinstance(names, str):
                names = names.split(',')
            for requested_prop in names:
                if not isinstance(requested_prop, str):
                    raise TypeError(f"property name must be a string. Given type {type(requested_prop)}")
                if not isinstance(self.properties[requested_prop], Property) or self.properties[requested_prop]._remote_info is None:
                    raise AttributeError("this property is not remote accessible")
                data[requested_prop] = self.properties[requested_prop].__get__(self, type(self))
        elif len(kwargs.keys()) != 0:
            for rename, requested_prop in kwargs.items():
                if not isinstance(self.properties[requested_prop], Property) or self.properties[requested_prop]._remote_info is None:
                    raise AttributeError("this property is not remote accessible")
                data[rename] = self.properties[requested_prop].__get__(self, type(self))                   
        return data 
    
    @action(URL_path='/properties', http_method=[HTTP_METHODS.PUT, HTTP_METHODS.PATCH])
    def _set_properties(self, **values : typing.Dict[str, typing.Any]) -> None:
        """ 
        set properties whose name is specified by keys of a dictionary
        
        Parameters
        ----------
        values: Dict[str, Any]
            dictionary of property names and its values
        """
        produced_error = False
        errors = ''
        for name, value in values.items():
            try:
                setattr(self, name, value)
            except Exception as ex:
                self.logger.error(f"could not set attribute {name} due to error {str(ex)}")
                errors += f'{name} : {str(ex)}\n'
                produced_error = True
        if produced_error:
            ex = RuntimeError("Some properties could not be set due to errors. " + 
                            "Check exception notes or server logs for more information.")
            ex.__notes__ = errors
            raise ex from None

    @action(URL_path='/properties/db', http_method=HTTP_METHODS.GET)     
    def _get_properties_in_db(self) -> typing.Dict[str, JSONSerializable]:
        """
        get all properties in the database
        
        Returns
        -------
        Dict[str, JSONSerializable]
            dictionary of property names and their values
        """
        if not hasattr(self, 'db_engine'):
            return {}
        props = self.db_engine.get_all_properties()
        final_list = {}
        for name, prop in props.items():
            try:
                self.http_serializer.dumps(prop)
                final_list[name] = prop
            except Exception as ex:
                self.logger.error(f"could not serialize property {name} to JSON due to error {str(ex)}, skipping this property")
        return final_list

    @action(URL_path='/properties', http_method=HTTP_METHODS.POST)
    def _add_property(self, name : str, prop : JSON) -> None:
        """
        add a property to the object
        
        Parameters
        ----------
        name: str
            name of the property
        prop: Property
            property object
        """
        raise NotImplementedError("this method will be implemented properly in a future release")
        prop = Property(**prop)
        self.properties.add(name, prop)
        self._prepare_resources()
        # instruct the clients to fetch the new resources

    @property
    def event_publisher(self) -> EventPublisher:
        """
        event publishing PUB socket owning object, valid only after 
        ``run()`` is called, otherwise raises AttributeError.
        """
        return self._event_publisher 
                   
    @event_publisher.setter
    def event_publisher(self, value : EventPublisher) -> None:
        if self._event_publisher is not None:
            raise AttributeError("Can set event publisher only once")
        
        def recusively_set_event_publisher(obj : Thing, publisher : EventPublisher) -> None:
            for name, evt in inspect._getmembers(obj, lambda o: isinstance(o, Event), getattr_without_descriptor_read):
                assert isinstance(evt, Event), "object is not an event"
                # above is type definition
                e = evt.__get__(obj, type(obj)) 
                e.publisher = publisher 
                e._remote_info.socket_address = publisher.socket_address
                self.logger.info(f"registered event '{evt.friendly_name}' serving at PUB socket with address : {publisher.socket_address}")
            for name, subobj in inspect._getmembers(obj, lambda o: isinstance(o, Thing), getattr_without_descriptor_read):
                if name == '_owner':
                    continue 
                recusively_set_event_publisher(subobj, publisher)
            obj._event_publisher = publisher            

        recusively_set_event_publisher(self, value)


    @action(URL_path='/properties/db-reload', http_method=HTTP_METHODS.POST)
    def load_properties_from_DB(self):
        """
        Load and apply property values which have ``db_init`` or ``db_persist``
        set to ``True`` from database
        """
        if not hasattr(self, 'db_engine'):
            return
        missing_properties = self.db_engine.create_missing_properties(self.__class__.properties.db_init_objects,
                                                                    get_missing_property_names=True)
        # 4. read db_init and db_persist objects
        with edit_constant_parameters(self):
            for db_prop, value in self.db_engine.get_all_properties().items():
                try:
                    prop_desc = self.properties.descriptors[db_prop]
                    if (prop_desc.db_init or prop_desc.db_persist) and db_prop not in missing_properties:
                        setattr(self, db_prop, value) # type: ignore
                except Exception as ex:
                    self.logger.error(f"could not set attribute {db_prop} due to error {str(ex)}")

        
    @action(URL_path='/resources/postman-collection', http_method=HTTP_METHODS.GET)
    def get_postman_collection(self, domain_prefix : str = None):
        """
        organised postman collection for this object
        """
        from .api_platforms import postman_collection
        return postman_collection.build(instance=self, 
                    domain_prefix=domain_prefix if domain_prefix is not None else self._object_info.http_server)
    

    @action(URL_path='/resources/wot-td', http_method=HTTP_METHODS.GET)
    def get_thing_description(self, authority : typing.Optional[str] = None, ignore_errors : bool = False): 
                            # allow_loose_schema : typing.Optional[bool] = False): 
        """
        generate thing description schema of Web of Things https://www.w3.org/TR/wot-thing-description11/.
        one can use the node-wot as a client for the object with the generated schema 
        (https://github.com/eclipse-thingweb/node-wot). Other WoT related tools based on TD will be compatible. 
        Composed Things that are not the top level object is currently not supported.
        
        Parameters
        ----------
        authority: str, optional
            protocol with DNS or protocol with hostname+port, for example 'https://my-pc:8080' or 
            'http://my-pc:9090' or 'https://IT-given-domain-name'. If absent, a value will be automatically
            given using ``socket.gethostname()`` and the port at which the last HTTPServer (``hololinked.server.HTTPServer``) 
            attached to this object was running.
        ignore_errors: bool, optional, Default False
            if True, offending interaction affordances will be removed from the schema. This is useful to build partial but working
            schema always.             
        Returns
        -------
        hololinked.wot.td.ThingDescription
            represented as an object in python, gets automatically serialized to JSON when pushed out of the socket. 
        """
        # allow_loose_schema: bool, optional, Default False 
        #     Experimental properties, actions or events for which schema was not given will be supplied with a suitable 
        #     value for node-wot to ignore validation or claim the accessed value for complaint with the schema.
        #     In other words, schema validation will always pass.  
        from .td import ThingDescription
        return ThingDescription(instance=self, authority=authority or self._object_info.http_server,
                                    allow_loose_schema=False, ignore_errors=ignore_errors).produce() #allow_loose_schema)   


    @action(URL_path='/exit', http_method=HTTP_METHODS.POST)                                                                                                                                          
    def exit(self) -> None:
        """
        Exit the object without killing the eventloop that runs this object. If Thing was 
        started using the run() method, the eventloop is also killed. This method can
        only be called remotely.
        """
        if self.rpc_server is None:
            return 
        if self._owner is None:
            self.rpc_server.stop_polling()
            raise BreakInnerLoop # stops the inner loop of the object
        else:
            warnings.warn("call exit on the top object, composed objects cannot exit the loop.", RuntimeWarning)
    
    @action()
    def ping(self) -> None:
        """ping the Thing to see if it is alive"""
        pass 

    def run(self, 
            zmq_protocols : typing.Union[typing.Sequence[ZMQ_PROTOCOLS], 
                                         ZMQ_PROTOCOLS] = ZMQ_PROTOCOLS.IPC, 
            # expose_eventloop : bool = False,
            **kwargs 
        ) -> None:
        """
        Quick-start ``Thing`` server by creating a default eventloop & ZMQ servers. This 
        method is blocking until exit() is called.

        Parameters
        ----------
        zmq_protocols: Sequence[ZMQ_PROTOCOLS] | ZMQ_Protocools, Default ZMQ_PROTOCOLS.IPC or "IPC"
            zmq transport layers at which the object is exposed. 
            TCP - provides network access apart from HTTP - please supply a socket address additionally.  
            IPC - inter process communication - connection can be made from other processes running 
            locally within same computer. No client on the network will be able to contact the object using
            this transport. INPROC - one main python process spawns several threads in one of which the ``Thing``
            the running. The object can be contacted by a client on another thread but neither from other processes 
            or the network. One may use more than one form of transport.  All requests made will be anyway queued internally
            irrespective of origin. 
        
        **kwargs
            tcp_socket_address: str, optional
                socket_address for TCP access, for example: tcp://0.0.0.0:61234
            context: zmq.asyncio.Context, optional
                zmq context to be used. If not supplied, a new context is created.
                For INPROC clients, you need to provide a context.
        """
        # expose_eventloop: bool, False
        #     expose the associated Eventloop which executes the object. This is generally useful for remotely 
        #     adding more objects to the same event loop.
        # dont specify http server as a kwarg, as the other method run_with_http_server has to be used
        context = kwargs.get('context', None)
        if context is not None and not isinstance(context, zmq.asyncio.Context):
            raise TypeError("context must be an instance of zmq.asyncio.Context")
        context = context or zmq.asyncio.Context()

        self.rpc_server = RPCServer(
                                instance_name=self.instance_name, 
                                server_type=self.__server_type__.value, 
                                context=context, 
                                protocols=zmq_protocols, 
                                zmq_serializer=self.zmq_serializer, 
                                http_serializer=self.http_serializer, 
                                tcp_socket_address=kwargs.get('tcp_socket_address', None),
                                logger=self.logger
                            ) 
        self.message_broker = self.rpc_server.inner_inproc_server
        self.event_publisher = self.rpc_server.event_publisher 

        from .eventloop import EventLoop
        self.event_loop = EventLoop(
                    instance_name=f'{self.instance_name}/eventloop', 
                    things=[self], 
                    logger=self.logger,
                    zmq_serializer=self.zmq_serializer, 
                    http_serializer=self.http_serializer, 
                    expose=False, # expose_eventloop
                )
        
        if kwargs.get('http_server', None):
            from .HTTPServer import HTTPServer
            httpserver = kwargs.pop('http_server')
            assert isinstance(httpserver, HTTPServer)
            httpserver._zmq_protocol = ZMQ_PROTOCOLS.INPROC
            httpserver._zmq_inproc_socket_context = context
            httpserver._zmq_inproc_event_context = self.event_publisher.context
            assert httpserver.all_ok
            httpserver.tornado_instance.listen(port=httpserver.port, address=httpserver.address)
        self.event_loop.run()


    def run_with_http_server(self, port : int = 8080, address : str = '0.0.0.0', 
                # host : str = None, 
                allowed_clients : typing.Union[str, typing.Iterable[str]] = None,   
                ssl_context : ssl.SSLContext = None, # protocol_version : int = 1, 
                # network_interface : str = 'Ethernet', 
                **kwargs):
        """
        Quick-start ``Thing`` server by creating a default eventloop & servers. This 
        method is fully blocking.

        Parameters
        ----------
        port: int
            the port at which the HTTP server should be run (unique)
        address: str
            set custom IP address, default is localhost (0.0.0.0)
        ssl_context: ssl.SSLContext | None
            use it for highly customized SSL context to provide encrypted communication
        allowed_clients
            serves request and sets CORS only from these clients, other clients are rejected with 403. Unlike pure CORS
            feature, the server resource is not even executed if the client is not an allowed client.
        **kwargs,
            certfile: str
                alternative to SSL context, provide certificate file & key file to allow the server to create a SSL connection on its own
            keyfile: str
                alternative to SSL context, provide certificate file & key file to allow the server to create a SSL connection on its own
            request_handler: RPCHandler
                custom web request handler of your choice
            event_handler: BaseHandler | EventHandler
                custom event handler of your choice for handling events
        """
        # network_interface: str
        #     Currently there is no logic to detect the IP addresss (as externally visible) correctly, therefore please 
        #     send the network interface name to retrieve the IP. If a DNS server is present, you may leave this field
        # host: str
        #     Host Server to subscribe to coordinate starting sequence of things & web GUI
        
        from .HTTPServer import HTTPServer
        
        http_server = HTTPServer(
            [self.instance_name], logger=self.logger, serializer=self.http_serializer, 
            port=port, address=address, ssl_context=ssl_context,
            allowed_clients=allowed_clients, schema_validator=self.schema_validator,
            # network_interface=network_interface, 
            **kwargs,
        )
        
        self.run(
            zmq_protocols=ZMQ_PROTOCOLS.INPROC,
            http_server=http_server,
            context=kwargs.get('context', None)
        ) # blocks until exit is called

        http_server.tornado_instance.stop()

       






import sys
import logging
import re 
import asyncio
import inspect
import typing
import asyncio
import types
import traceback
import typing
import ifaddr



def get_IP_from_interface(interface_name : str = 'Ethernet', adapter_name = None) -> str:
    """
    Get IP address of specified interface. Generally necessary when connected to the network
    through multiple adapters and a server binds to only one adapter at a time. 

    Parameters
    ----------
    interface_name: str
        Ethernet, Wifi etc. 
    adapter_name: optional, str
        name of the adapter if available

    Returns
    -------
    str: 
        IP address of the interface
    """
    adapters = ifaddr.get_adapters(include_unconfigured=True)
    for adapter in adapters:
        if not adapter_name:
            for ip in adapter.ips:
                if interface_name == ip.nice_name:
                    if ip.is_IPv4:
                        return ip.ip
        elif adapter_name == adapter.nice_name:
            for ip in adapter.ips:
                if interface_name == ip.nice_name:
                    if ip.is_IPv4:
                        return ip.ip
    raise ValueError(f"interface name {interface_name} not found in system interfaces.")
            

def format_exception_as_json(exc : Exception) -> typing.Dict[str, typing.Any]: 
    """
    return exception as a JSON serializable dictionary
    """
    return {
        "message" : str(exc),
        "type" : repr(exc).split('(', 1)[0],
        "traceback" : traceback.format_exc().splitlines(),
        "notes" : exc.__notes__ if hasattr(exc, "__notes__") else None 
    }


def pep8_to_URL_path(word : str) -> str: 
    """
    Make an underscored, lowercase form from the expression in the string.
    Example::
        >>> pep8_to_dashed_URL("device_type")
        'device-type'
    """
    val = re.sub(r'_+', '-', word.lstrip('_').rstrip('_'))
    return val.replace(' ', '-')


def get_default_logger(name : str, log_level : int = logging.INFO, log_file = None,
                format : str = '%(levelname)-8s - %(asctime)s:%(msecs)03d - %(name)s - %(message)s' ) -> logging.Logger:
    """
    the default logger used by most of hololinked package, when arguments are not modified.
    StreamHandler is always created, pass log_file for a FileHandler as well.

    Parameters
    ----------
    name: str 
        name of logger
    log_level: int 
        log level
    log_file: str
        valid path to file
    format: str
        log format

    Returns
    -------
    logging.Logger:
        created logger
    """
    logger = logging.getLogger(name) 
    logger.setLevel(log_level)
    default_handler = logging.StreamHandler(sys.stdout)
    default_handler.setFormatter(logging.Formatter(format, datefmt='%Y-%m-%dT%H:%M:%S'))
    logger.addHandler(default_handler)
    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(logging.Formatter(format, datefmt='%Y-%m-%dT%H:%M:%S'))
        logger.addHandler(file_handler)
    return logger


def run_coro_sync(coro : typing.Coroutine):
    """
    run coroutine synchronously
    """
    try:
        eventloop = asyncio.get_event_loop()
    except RuntimeError:
        eventloop = asyncio.new_event_loop()
    if eventloop.is_running():
        raise RuntimeError(f"asyncio event loop is already running, cannot setup coroutine {coro.__name__} to run sync, please await it.")
        # not the same as RuntimeError catch above.  
    else:
        return eventloop.run_until_complete(coro)


def run_callable_somehow(method : typing.Union[typing.Callable, typing.Coroutine]) -> typing.Any:
    """
    run method if synchronous, or when async, either schedule a coroutine or run it until its complete
    """
    if not (asyncio.iscoroutinefunction(method) or asyncio.iscoroutine(method)):
        return method()
    try:
        eventloop = asyncio.get_event_loop()
    except RuntimeError:
        eventloop = asyncio.new_event_loop()
    if asyncio.iscoroutinefunction(method):
        coro = method()
    else:
        coro = method
    if eventloop.is_running():    
        # task =  # check later if lambda is necessary
        eventloop.create_task(coro)
    else:
        # task = method
        return eventloop.run_until_complete(coro)


def get_signature(callable : typing.Callable) -> typing.Tuple[typing.List[str], typing.List[type]]: 
    """
    Retrieve the names and types of arguments based on annotations for the given callable.

    Parameters
    ----------
    callable: Callable
        function or method (not tested with __call__)
    
    Returns
    -------
    tuple: List[str], List[type]
        arguments name and types respectively
    """
    arg_names = []
    arg_types = []

    for param in inspect.signature(callable).parameters.values():
        arg_name = param.name
        arg_type = param.annotation if param.annotation != inspect.Parameter.empty else None

        arg_names.append(arg_name)
        arg_types.append(arg_type)

    return arg_names, arg_types


def getattr_without_descriptor_read(instance, key):
    """
    supply to inspect._get_members (not inspect.get_members) to avoid calling 
    __get__ on hardware attributes
    """
    if key in instance.__dict__:
        return instance.__dict__[key]
    mro =  mro = (instance.__class__,) + inspect.getmro(instance.__class__)
    for base in mro:
        if key in base.__dict__:
            value = base.__dict__[key]
            if isinstance(value, types.FunctionType):
                method = getattr(instance, key, None)
                if isinstance(method, types.MethodType):
                    return method 
            return value 
    # for descriptor, first try to find it in class dict or instance dict (for instance descriptors (per_instance_descriptor=True))
    # and then getattr from the instance. For descriptors/property, it will be mostly at above two levels.
    return getattr(instance, key, None) # we can deal with None where we use this getter, so dont raise AttributeError  


def isclassmethod(method):
    """https://stackoverflow.com/questions/19227724/check-if-a-function-uses-classmethod"""
    bound_to = getattr(method, '__self__', None)
    if not isinstance(bound_to, type):
        # must be bound to a class
        return False
    name = method.__name__
    for cls in bound_to.__mro__:
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False


def issubklass(obj, cls):
    """
    Safely check if `obj` is a subclass of `cls`.

    Parameters:
    obj: The object to check if it's a subclass.
    cls: The class (or tuple of classes) to compare against.

    Returns:
    bool: True if `obj` is a subclass of `cls`, False otherwise.
    """
    try:
        # Check if obj is a class or a tuple of classes
        if isinstance(obj, type):
            return issubclass(obj, cls)
        elif isinstance(obj, tuple):
            # Ensure all elements in the tuple are classes
            return all(isinstance(o, type) for o in obj) and issubclass(obj, cls)
        else:
            return False
    except TypeError:
        return False
  

__all__ = [
    get_IP_from_interface.__name__,
    format_exception_as_json.__name__,
    pep8_to_URL_path.__name__,
    get_default_logger.__name__,
    run_coro_sync.__name__,
    run_callable_somehow.__name__,
    get_signature.__name__,
    isclassmethod.__name__,
    issubklass.__name__
]






import asyncio
import typing
import unittest
import logging
import multiprocessing
from hololinked.server.dataklasses import ActionInfoValidator
from hololinked.server.thing import Thing, action
from hololinked.server.utils import isclassmethod
from hololinked.param import ParameterizedFunction
from hololinked.client import ObjectProxy
from hololinked.server.properties import Number, String, ClassSelector
try:
    from .utils import TestCase, TestRunner
    from .things import TestThing
except ImportError:
    from utils import TestCase, TestRunner
    from things import start_thing_forked 



class TestThing(Thing):

    def action_echo(self, value):
        return value

    @classmethod
    def action_echo_with_classmethod(self, value):
        return value
    
    async def action_echo_async(self, value):
        await asyncio.sleep(0.1)
        return value
    
    @classmethod
    async def action_echo_async_with_classmethod(self, value):
        await asyncio.sleep(0.1)
        return value
    
    class typed_action(ParameterizedFunction):

        arg1 = Number(bounds=(0, 10), step=0.5, default=5, crop_to_bounds=True, 
                    doc='arg1 description')
        arg2 = String(default='hello', doc='arg2 description', regex='[a-z]+')
        arg3 = ClassSelector(class_=(int, float, str),
                            default=5, doc='arg3 description')

        def __call__(self, instance, arg1, arg2, arg3):
            return instance.instance_name, arg1, arg2, arg3


    class typed_action_without_call(ParameterizedFunction):

        arg1 = Number(bounds=(0, 10), step=0.5, default=5, crop_to_bounds=True, 
                    doc='arg1 description')
        arg2 = String(default='hello', doc='arg2 description', regex='[a-z]+')
        arg3 = ClassSelector(class_=(int, float, str),
                            default=5, doc='arg3 description')


    class typed_action_async(ParameterizedFunction):
            
        arg1 = Number(bounds=(0, 10), step=0.5, default=5, crop_to_bounds=True, 
                    doc='arg1 description')
        arg2 = String(default='hello', doc='arg2 description', regex='[a-z]+')
        arg3 = ClassSelector(class_=(int, float, str),
                            default=5, doc='arg3 description')

        async def __call__(self, instance, arg1, arg2, arg3):
            await asyncio.sleep(0.1)
            return instance.instance_name, arg1, arg2, arg3


    def __internal__(self, value):
        return value

    def incorrectly_decorated_method(self, value):
        return value
    
    def not_an_action(self, value):
        return value
    
    async def not_an_async_action(self, value):
        await asyncio.sleep(0.1)
        return value
    


class TestAction(TestCase):

    @classmethod
    def setUpClass(self):
        print("test action")
        self.thing_cls = TestThing 

    @classmethod
    def tearDownClass(self) -> None:
        print("tear down test action")

    def test_1_allowed_actions(self):
        # instance method can be decorated with action
        self.assertEqual(self.thing_cls.action_echo, action()(self.thing_cls.action_echo))
        # classmethod can be decorated with action
        self.assertEqual(self.thing_cls.action_echo_with_classmethod, 
                        action()(self.thing_cls.action_echo_with_classmethod))
        self.assertTrue(isclassmethod(self.thing_cls.action_echo_with_classmethod))
        # async methods can be decorated with action    
        self.assertEqual(self.thing_cls.action_echo_async, 
                        action()(self.thing_cls.action_echo_async))
        # async classmethods can be decorated with action
        self.assertEqual(self.thing_cls.action_echo_async_with_classmethod, 
                        action()(self.thing_cls.action_echo_async_with_classmethod))
        self.assertTrue(isclassmethod(self.thing_cls.action_echo_async_with_classmethod))
        # parameterized function can be decorated with action
        self.assertEqual(self.thing_cls.typed_action, action(safe=True)(self.thing_cls.typed_action))
        self.assertEqual(self.thing_cls.typed_action_without_call, action(idempotent=True)(self.thing_cls.typed_action_without_call))
        self.assertEqual(self.thing_cls.typed_action_async, action(synchronous=True)(self.thing_cls.typed_action_async))
       
        
    def test_2_remote_info(self):
        # basic check if the remote_info is correct, although this test is not necessary, not recommended and 
        # neither particularly useful
        remote_info = self.thing_cls.action_echo._remote_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator) # type definition
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.isproperty)
        self.assertFalse(remote_info.isparameterized)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.safe)  
        self.assertFalse(remote_info.idempotent)
        self.assertFalse(remote_info.synchronous)

        remote_info = self.thing_cls.action_echo_async._remote_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator) # type definition
        self.assertTrue(remote_info.isaction)
        self.assertTrue(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertFalse(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertFalse(remote_info.synchronous)

        remote_info = self.thing_cls.action_echo_with_classmethod._remote_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator) # type definition
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertFalse(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertFalse(remote_info.synchronous)

        remote_info = self.thing_cls.typed_action._remote_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertTrue(remote_info.isparameterized)
        self.assertTrue(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertFalse(remote_info.synchronous)

        remote_info = self.thing_cls.typed_action_without_call._remote_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)
        self.assertTrue(remote_info.isaction)
        self.assertFalse(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertTrue(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertTrue(remote_info.idempotent)
        self.assertFalse(remote_info.synchronous)

        remote_info = self.thing_cls.typed_action_async._remote_info
        self.assertIsInstance(remote_info, ActionInfoValidator)
        assert isinstance(remote_info, ActionInfoValidator)
        self.assertTrue(remote_info.isaction)
        self.assertTrue(remote_info.iscoroutine)
        self.assertFalse(remote_info.isproperty)
        self.assertTrue(remote_info.isparameterized)
        self.assertFalse(remote_info.safe)
        self.assertFalse(remote_info.idempotent)
        self.assertTrue(remote_info.synchronous)


    def test_3_api_and_invalid_actions(self):
        # done allow action decorator to be terminated without '()' on a method
        with self.assertRaises(TypeError) as ex:
           action(self.thing_cls.incorrectly_decorated_method)
        self.assertTrue(str(ex.exception).startswith("URL_path should be a string, not a function/method, did you decorate"))
        
        # dunder methods cannot be decorated with action
        with self.assertRaises(ValueError) as ex:
            action()(self.thing_cls.__internal__)
        self.assertTrue(str(ex.exception).startswith("dunder objects cannot become remote"))

        # only functions and methods can be decorated with action
        for obj in [self.thing_cls, str, 1, 1.0, 'Str', True, None, object(), type, property]:
            with self.assertRaises(TypeError) as ex:
                action()(obj) # not an action
            self.assertTrue(str(ex.exception).startswith("target for action or is not a function/method."))

        with self.assertRaises(ValueError) as ex:
            action(safe=True, some_kw=1)
        self.assertTrue(str(ex.exception).startswith("Only 'safe', 'idempotent', 'synchronous' are allowed"))
            

    def test_4_exposed_actions(self):
        self.assertTrue(hasattr(self.thing_cls.action_echo, '_remote_info'))
        done_queue = multiprocessing.Queue()
        start_thing_forked(self.thing_cls, instance_name='test-action', done_queue=done_queue,
                                        log_level=logging.ERROR+10, prerun_callback=expose_actions)

        thing_client = ObjectProxy('test-action', log_level=logging.ERROR) # type: TestThing

        self.assertTrue(thing_client.action_echo(1) == 1)
        self.assertTrue(thing_client.action_echo_async("string") == "string")
        self.assertTrue(thing_client.typed_action(arg1=1, arg2='hello', arg3=5) == ['test-action', 1, 'hello', 5])
        self.assertTrue(thing_client.typed_action_async(arg1=2.5, arg2='hello', arg3='foo') == ['test-action', 2.5, 'hello', 'foo'])

        with self.assertRaises(NotImplementedError) as ex:
            thing_client.typed_action_without_call(arg1=1, arg2='hello', arg3=5), 
        self.assertTrue(str(ex.exception).startswith("Subclasses must implement __call__"))
        
        with self.assertRaises(AttributeError) as ex:
            thing_client.__internal__(1)
        self.assertTrue(str(ex.exception).startswith("'ObjectProxy' object has no attribute '__internal__'"))

        with self.assertRaises(AttributeError) as ex:
            thing_client.not_an_action("foo")
        self.assertTrue(str(ex.exception).startswith("'ObjectProxy' object has no attribute 'not_an_action'"))

        with self.assertRaises(AttributeError) as ex:
            thing_client.not_an_async_action(1)
        self.assertTrue(str(ex.exception).startswith("'ObjectProxy' object has no attribute 'not_an_async_action'"))

        thing_client.exit()

        self.assertTrue(done_queue.get() == 'test-action')



def expose_actions(thing_cls):
    action()(thing_cls.action_echo)
        # classmethod can be decorated with action   
    action()(thing_cls.action_echo_with_classmethod)   
    # async methods can be decorated with action       
    action()(thing_cls.action_echo_async)
    # async classmethods can be decorated with action    
    action()(thing_cls.action_echo_async_with_classmethod)
    # parameterized function can be decorated with action
    action(safe=True)(thing_cls.typed_action)
    action(idempotent=True)(thing_cls.typed_action_without_call)
    action(synchronous=True)(thing_cls.typed_action_async)



if __name__ == '__main__':
    unittest.main(testRunner=TestRunner())


# Order of import is reflected in this file to avoid circular imports
from .constants import *
from .serializers import *
from .config import *
from .zmq_message_brokers import *
from .events import *
from .action import *
from .property import *
from .database import *
from .thing import *
from .eventloop import *
from .HTTPServer import *





import glob
import re
import os.path
import datetime as dt
import typing 
import numbers
import sys
import collections.abc
from collections import OrderedDict

from .utils import *
from .exceptions import *
from .parameterized import ParamOverrides, Parameterized, ParameterizedFunction, descendents, dt_types, Parameter



class Infinity(object):
    """
    An instance of this class represents an infinite value. Unlike
    Python's float('inf') value, this object can be safely compared
    with gmpy numeric types across different gmpy versions.

    All operators on Infinity() return Infinity(), apart from the
    comparison and equality operators. Equality works by checking
    whether the two objects are both instances of this class.
    """
    def __eq__  (self, other): return isinstance(other,self.__class__)
    def __ne__  (self, other): return not self == other
    def __lt__  (self, other): return False
    def __le__  (self, other): return False
    def __gt__  (self, other): return True
    def __ge__  (self, other): return True
    def __add__ (self, other): return self
    def __radd__(self, other): return self
    def __ladd__(self, other): return self
    def __sub__ (self, other): return self
    def __iadd_ (self, other): return self
    def __isub__(self, other): return self
    def __repr__(self):       return "Infinity()"
    def __str__ (self):       return repr(self)


 
class String(Parameter):
    """
    A string parameter with a default value and optional regular expression (regex) matching.

    Example of using a regex to implement IPv4 address matching::

      class IPAddress(String):
        '''IPv4 address as a string (dotted decimal notation)'''
       def __init__(self, default="0.0.0.0", allow_None=False, **kwargs):
           ip_regex = r'^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$'
           super(IPAddress, self).__init__(default=default, regex=ip_regex, **kwargs)
    """

    __slots__ = ['regex']

    def __init__(self, default : typing.Optional[str] = "", *, regex : typing.Optional[str] = None, 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False, 
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, class_member : bool = False, 
            fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None, 
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, doc=doc, constant=constant, readonly=readonly, 
            allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel,
            precedence=precedence)
        self.regex = regex
    
    def validate_and_adapt(self, value : typing.Any) -> str: 
        self._assert(value, self.regex, self.allow_None)
        return value

    @classmethod
    def _assert(obj, value : typing.Any, regex : typing.Optional[str] = None, allow_None : bool = False) -> None: 
        """
        the method that implements the validator
        """
        if value is None: 
            if allow_None:
                return 
            else:
                raise_ValueError(f"None not allowed for string type", obj)
        if not isinstance(value, str):
            raise_TypeError("given value is not string type, but {}.".format(type(value)), obj)
        if regex is not None:
            match = re.match(regex, value) 
            if match is None or match.group(0) != value:
                # match should be original string, not some substring
                raise_ValueError("given string value {} does not match regex {}.".format(value, regex), obj)

    @classmethod
    def isinstance(cls, value : typing.Any, regex : typing.Optional[str] = None, allow_None : bool = False) -> bool:
        """
        verify if given value is a string confirming to regex. 
        
        Args:
            value (Any): input value
            regex (str, None): regex required to match, leave None if unnecessary
            allow_None (bool): set True if None is tolerated

        Returns:
            bool: True if conformant, else False. Any exceptions due to wrong inputs resulting in TypeError and ValueError
                also lead to False
        """
        try:
            cls._assert(value, regex, allow_None)
            return True
        except (TypeError, ValueError):
            return False

   

class Bytes(String):
    """
    A bytes parameter with a default value and optional regular
    expression (regex) matching.

    Similar to the string parameter, but instead of type basestring
    this parameter only allows objects of type bytes (e.g. b'bytes').
    """

    @classmethod
    def _assert(obj, value : typing.Any, regex : typing.Optional[bytes] = None, allow_None : bool = False) -> None:
        """
        verify if given value is a bytes confirming to regex. 
        
        Args:
            value (Any): input value
            regex (str, None): regex required to match, leave None if unnecessary
            allow_None (bool): set True if None is tolerated

        Raises:
            TypeError: if given type is not bytes
            ValueError: if regex does not match
        """
        if value is None: 
            if allow_None:
                return 
            else:
                raise_ValueError(f"None not allowed for string type", obj)
        if not isinstance(value, bytes):
            raise_TypeError("given value is not bytes type, but {}.".format(type(value)), obj)
        if regex is not None:
            match = re.match(regex, value) 
            if match is None or match.group(0) != value:
                # match should be original string, not some substring
                raise_ValueError("given bytes value {} does not match regex {}.".format(value, regex), obj)
    


class IPAddress(Parameter):

    __slots__ = ['allow_localhost', 'allow_ipv4', 'allow_ipv6']
    
    def __init__(self, default : typing.Optional[str] = "0.0.0.0", *, allow_ipv4 : bool = True, allow_ipv6 : bool = True, 
            allow_localhost : bool = True,
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, 
            allow_None : bool = False, per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, doc=doc, constant=constant, readonly=readonly, 
            allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)
        self.allow_localhost = allow_localhost 
        self.allow_ipv4 = allow_ipv4
        self.allow_ipv6 = allow_ipv6

    def validate_and_adapt(self, value: typing.Any) -> str:
        self._assert(value, self.allow_ipv4, self.allow_ipv6, self.allow_localhost, self.allow_None)
        return value
    
    @classmethod
    def _assert(obj, value : typing.Any, allow_ipv4 : bool = True, allow_ipv6 : bool = True,
                allow_localhost : bool = True, allow_None : bool = False) -> None:
        if value is None and allow_None:
            return
        if not isinstance(value, str):
            raise_TypeError('given value for IP address not a string, but type {}'.format(type(value)), obj)
        if allow_localhost and value == 'localhost':
            return
        if not ((allow_ipv4 and (obj.isipv4(value) or obj.isipv4cidr(value))) 
                        or (allow_ipv6 and (obj.isipv6(value) or obj.isipv6cidr(value)))):
            raise_ValueError("Given value {} is not a valid IP address.".format(value), obj)
        
    @classmethod
    def isinstance(obj, value : typing.Any, allow_ipv4 : bool = True, allow_ipv6 : bool = True ,
                allow_localhost : bool = True, allow_None : bool = False) -> bool:
        try:
            obj._assert(value, allow_ipv4, allow_ipv6, allow_localhost, allow_None)
            return True
        except (TypeError, ValueError):
            return False
        
    @classmethod
    def isipv4(obj, value : str) -> bool:
        """
        Return whether a given value is a valid IP version 4 address.

        This validator is based on `WTForms IPAddress validator`_

        .. _WTForms IPAddress validator:
        https://github.com/wtforms/wtforms/blob/master/wtforms/validators.py

        Args:
            value (str): IP address string to validate, other types will raise unexpected errors (mostly attribute error)
        
        Returns:
            bool : True if conformant
        """
        groups = value.split(".")
        if (
            len(groups) != 4
            or any(not x.isdigit() for x in groups)
            or any(len(x) > 3 for x in groups)
        ):
            return False
        return all(0 <= int(part) < 256 for part in groups)
        
    
    @classmethod
    def isipv4cidr(obj, value : str) -> bool:
        """
        Return whether a given value is a valid CIDR-notated IP version 4
        address range.

        This validator is based on RFC4632 3.1.

        Args:
            value (str): IP address string to validate, other types will raise unexpected errors (mostly attribute error)
        
        Returns:
            bool : True if conformant
        """
        try:
            prefix, suffix = value.split('/', 2)
        except ValueError:
            return False
        if not obj.is_ipv4(prefix) or not suffix.isdigit():
            return False
        return 0 <= int(suffix) <= 32
    
    @classmethod
    def isipv6(obj, value : str) -> bool:
        """
        Return whether a given value is a valid IP version 6 address
        (including IPv4-mapped IPv6 addresses).

        This validator is based on `WTForms IPAddress validator`_.

        .. _WTForms IPAddress validator:
        https://github.com/wtforms/wtforms/blob/master/wtforms/validators.py

        Examples::

            >>> ipv6('abcd:ef::42:1')
            True

            >>> ipv6('::ffff:192.0.2.128')
            True

            >>> ipv6('::192.0.2.128')
            True

            >>> ipv6('abc.0.0.1')
            ValidationFailure(func=ipv6, args={'value': 'abc.0.0.1'})

        .. versionadded:: 0.2

        :param value: IP address string to validate
        """
        ipv6_groups = value.split(':')
        if len(ipv6_groups) == 1:
            return False
        ipv4_groups = ipv6_groups[-1].split('.')

        if len(ipv4_groups) > 1:
            if not obj.is_ipv4(ipv6_groups[-1]):
                return False
            ipv6_groups = ipv6_groups[:-1]
        else:
            ipv4_groups = []

        count_blank = 0
        for part in ipv6_groups:
            if not part:
                count_blank += 1
                continue
            try:
                num = int(part, 16)
            except ValueError:
                return False
            else:
                if not 0 <= num <= 65536 or len(part) > 4:
                    return False

        max_groups = 6 if ipv4_groups else 8
        part_count = len(ipv6_groups) - count_blank
        if count_blank == 0 and part_count == max_groups:
            # no :: -> must have size of max_groups
            return True
        elif count_blank == 1 and ipv6_groups[-1] and ipv6_groups[0] and part_count < max_groups:
            # one :: inside the address or prefix or suffix : -> filter least two cases
            return True
        elif count_blank == 2 and part_count < max_groups and (
                ((ipv6_groups[0] and not ipv6_groups[-1]) or (not ipv6_groups[0] and ipv6_groups[-1])) or ipv4_groups):
            # leading or trailing :: or : at end and begin -> filter last case
            # Check if it has ipv4 groups because they get removed from the ipv6_groups
            return True
        elif count_blank == 3 and part_count == 0:
            # :: is the address -> filter everything else
            return True
        return False

    @classmethod
    def isipv6cidr(obj, value : str) -> bool:
        """
        Returns whether a given value is a valid CIDR-notated IP version 6
        address range.

        This validator is based on RFC4632 3.1.

        Examples::

            >>> ipv6_cidr('::1/128')
            True

            >>> ipv6_cidr('::1')
            ValidationFailure(func=ipv6_cidr, args={'value': '::1'})
        """
        try:
            prefix, suffix = value.split('/', 2)
        except ValueError:
            return False
        if not obj.is_ipv6(prefix) or not suffix.isdigit():
            return False
        return 0 <= int(suffix) <= 128



class Number(Parameter):
    """
    A numeric parameter with a default value and optional bounds.

    There are two types of bounds: ``bounds`` and
    ``softbounds``. ``bounds`` are hard bounds: the parameter must
    have a value within the specified range.  The default bounds are
    (None,None), meaning there are actually no hard bounds.  One or
    both bounds can be set by specifying a value
    (e.g. bounds=(None,10) means there is no lower bound, and an upper
    bound of 10). Bounds are inclusive by default, but exclusivity
    can be specified for each bound by setting inclusive_bounds
    (e.g. inclusive_bounds=(True,False) specifies an exclusive upper
    bound).

    Using a default value outside the hard
    bounds, or one that is not numeric, results in an exception.

    As a special case, if allow_None=True (which is true by default if
    the parameter has a default of None when declared) then a value
    of None is also allowed.

    A separate function set_in_bounds() is provided that will
    silently crop the given value into the legal range, for use
    in, for instance, a GUI.

    ``softbounds`` are present to indicate the typical range of
    the parameter, but are not enforced. Setting the soft bounds
    allows, for instance, a GUI to know what values to display on
    sliders for the Number.

    Example of creating a Number::
        AB = Number(default=0.5, bounds=(None,10), softbounds=(0,1), doc='Distance from A to B.')

    """

    __slots__ = ['bounds', 'inclusive_bounds', 'crop_to_bounds', 'dtype', 'step']

    def __init__(self, default : typing.Optional[typing.Union[float, int]] = 0.0, *, bounds : typing.Optional[typing.Tuple] = None, 
            crop_to_bounds : bool = False, inclusive_bounds : typing.Tuple = (True,True), step : typing.Any = None, 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, doc=doc, constant=constant, readonly=readonly, 
            allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, deepcopy_default=deepcopy_default, 
            class_member=class_member, fget=fget, fset=fset, fdel=fdel, precedence=precedence)
        self.bounds = bounds
        self.crop_to_bounds = crop_to_bounds
        self.inclusive_bounds = inclusive_bounds
        self.dtype = (float, int)
        self.step = step

    def set_in_bounds(self, obj : typing.Union[Parameterized, typing.Any], value : typing.Union[float, int]) -> None:
        """
        Set to the given value, but cropped to be within the legal bounds.
        See crop_to_bounds for details on how cropping is done.
        """
        self._assert(value, self.dtype, None, (False, False), self.allow_None)
        bounded_value = self._crop_to_bounds(value) 
        super().__set__(obj, bounded_value)

    def _crop_to_bounds(self, value : typing.Union[int, float]) -> typing.Union[int, float]:
        """
        Return the given value cropped to be within the hard bounds
        for this parameter.

        If a numeric value is passed in, check it is within the hard
        bounds. If it is larger than the high bound, return the high
        bound. If it's smaller, return the low bound. In either case, the
        returned value could be None.  If a non-numeric value is passed
        in, set to be the default value (which could be None).  In no
        case is an exception raised; all values are accepted.
        """
        # Values outside the bounds are silently cropped to
        # be inside the bounds.
        assert self.bounds is not None, "Cannot crop to bounds when bounds is None"
        vmin, vmax = self.bounds
        incmin, incmax = self.inclusive_bounds
        if vmin is not None:
            if value < vmin:
                if incmin: 
                    return vmin
                else: 
                    return vmin + self.step 
        if vmax is not None:
            if value > vmax:
                if incmax:
                    return vmax
                else:
                    return vmax - self.step
        return value
    
    def validate_and_adapt(self, value: typing.Any) -> typing.Union[int, float]:
        self._assert(value, self.dtype, None if self.crop_to_bounds else self.bounds, 
                    self.inclusive_bounds, self.allow_None)
        if self.crop_to_bounds and self.bounds and value is not None:
            return self._crop_to_bounds(value)
        return value
        
    @classmethod
    def _assert(obj, value, dtype : typing.Tuple, bounds : typing.Optional[typing.Tuple] = None, 
               inclusive_bounds : typing.Tuple[bool, bool] = (True, True), allow_None : bool = False):
        if allow_None and value is None:
            return
        if dtype is None:
            if not obj.isnumber(value):
                raise_TypeError("given value not of number type, but type {}.".format(type(value)), 
                                obj)
        elif not isinstance(value, dtype):
            raise_TypeError("given value not of type {}, but type {}.".format(dtype, type(value)), obj)
        if bounds:
            vmin, vmax = bounds
            incmin, incmax = inclusive_bounds   
            if vmax is not None:
                if incmax is True:
                    if not value <= vmax:
                        raise_ValueError("given value must be at most {}, not {}.".format(vmax, value), obj)
                else:
                    if not value < vmax:
                        raise_ValueError("Parameter must be less than {}, not {}.".format(vmax, value), obj)

            if vmin is not None:
                if incmin is True:
                    if not value >= vmin:
                        raise_ValueError("Parameter must be at least {}, not {}.".format(vmin, value), obj)
                else:
                    if not value > vmin:
                        raise_ValueError("Parameter must be greater than {}, not {}.".format(vmin, value), obj)
            return value 
    
    def _validate_step(self, value : typing.Any) -> None:
        if value is not None:
            if self.dtype: 
                if not isinstance(value, self.dtype):
                    raise_ValueError("Step can only be None or {}, not type {}.".format(self.dtype, type(value)), self)
            elif not self.isnumber(self.step):
                raise_ValueError("Step can only be None or numeric value, not type {}.".format(type(value)), self)

    def _post_slot_set(self, slot : str, old : typing.Any, value : typing.Any) -> None:
        if slot == 'step': 
            self._validate_step(value)
        return super()._post_slot_set(slot, old, value)

    @classmethod
    def isinstance(obj, value, dtype : typing.Tuple, bounds : typing.Optional[typing.Tuple] = None, 
               inclusive_bounds : typing.Tuple[bool, bool] = (True, True), allow_None : bool = False):
        try:
            obj._assert(value, dtype, bounds, inclusive_bounds, allow_None)
            return True
        except (ValueError, TypeError):
            return False 
    	
    @classmethod
    def isnumber(cls, value : typing.Any) -> bool:
        if isinstance(value, numbers.Number): return True
        # The extra check is for classes that behave like numbers, such as those
        # found in numpy, gmpy, etc.
        elif (hasattr(value, '__int__') and hasattr(value, '__add__')): return True
        # This is for older versions of gmpy
        elif hasattr(value, 'qdiv'): return True
        else: return False



class Integer(Number):
    """Numeric Parameter required to be an Integer"""

    def __init__(self, default : typing.Optional[int] = 0, *, bounds : typing.Optional[typing.Tuple] = None, 
            crop_to_bounds : bool = False, inclusive_bounds : typing.Tuple = (True,True), step : typing.Any = None, 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, class_member : bool = False, 
            fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, bounds=bounds, crop_to_bounds=crop_to_bounds, inclusive_bounds=inclusive_bounds, 
            doc=doc, constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)
        self.dtype = (int,)

    def _validate_step(self, step : int):
        if step is not None and not isinstance(step, int):
            raise_ValueError("Step can only be None or an integer value, not type {}".format(type(step)), self)



class Boolean(Parameter):
    """Binary or tristate Boolean Parameter."""

    def __init__(self, default : typing.Optional[bool] = False, *, 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, doc=doc, constant=constant, readonly=readonly, 
            allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, deepcopy_default=deepcopy_default, 
            class_member=class_member, fget=fget, fset=fset, fdel=fdel, precedence=precedence)

    def validate_and_adapt(self, value : typing.Any) -> bool:
        if self.allow_None and value is None:
            return 
        elif not isinstance(value, bool): 
            raise_ValueError("given value not boolean type, but type {}".format(type(value)), self)
        return value



class Iterable(Parameter):
    """A tuple or list Parameter (e.g. ('a',7.6,[3,5])) with a fixed tuple length."""

    __slots__ = ['bounds', 'length', 'item_type', 'dtype']

    def __init__(self, default : typing.Any, *, bounds : typing.Optional[typing.Tuple[int, int]] = None, 
            length : typing.Optional[int] = None, item_type : typing.Optional[typing.Tuple] = None, deepcopy_default : bool = False, 
            allow_None : bool = False, doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, 
            per_instance_descriptor : bool = False,  class_member : bool = False, 
            fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        """
        Initialize a tuple parameter with a fixed length (number of
        elements).  The length is determined by the initial default
        value, if any, and must be supplied explicitly otherwise.  The
        length is not allowed to change after instantiation.
        """
        super().__init__(default=default, doc=doc, constant=constant, readonly=readonly, 
            allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, deepcopy_default=deepcopy_default, 
            class_member=class_member, fget=fget, fset=fset, fdel=fdel, precedence=precedence)
        self.bounds = bounds 
        self.length = length
        self.item_type = item_type
        self.dtype = (list, tuple)

    def validate_and_adapt(self, value: typing.Any) -> typing.Union[typing.List, typing.Tuple]:
        self._assert(value, self.bounds, self.length, self.dtype, self.item_type, self.allow_None)
        return value
        
    @classmethod
    def _assert(obj, value : typing.Any, bounds : typing.Optional[typing.Tuple[int, int]] = None, 
            length : typing.Optional[int] = None, dtype :  typing.Union[type, typing.Tuple] = (list, tuple), 
            item_type : typing.Any = None, allow_None : bool = False) -> None:
        if value is None and allow_None:
            return
        if not isinstance(value, dtype):
            raise_ValueError("given value not of iterable type {}, but {}.".format(dtype, type(value)), obj)
        if bounds is not None:
            if not (len(value) >= bounds[0] and len(value) <= bounds[1]):
                raise_ValueError("given iterable is not of the correct length ({} instead of between {} and {}).".format(
                                len(value), 0 if not bounds[0] else bounds[0], bounds[1]), obj) 
        elif length is not None and len(value) != length:
            raise_ValueError("given iterable is not of correct length ({} instead of {})".format(len(value), length), 
                             obj)
        if item_type is not None: 
            for val in value:
                if not isinstance(val, item_type):
                    raise_TypeError("not all elements of given iterable of item type {}, found object of type {}".format(
                        item_type, type(val)), obj)

    @classmethod
    def isinstance(obj, value : typing.Any, bounds : typing.Optional[typing.Tuple[int, int]], 
                length : typing.Optional[int] = None, dtype : typing.Union[type, typing.Tuple] = (list, tuple), 
                item_type : typing.Any = None, allow_None : bool = False) -> bool:
        try: 
            obj._assert(value, bounds, length, dtype, item_type, allow_None)
            return True 
        except (ValueError, TypeError):
            return False



class Tuple(Iterable):

    __slots__ = ['accept_list', 'accept_item']

    def __init__(self, default : typing.Any, *, bounds : typing.Optional[typing.Tuple[int, int]] = None, 
            length: typing.Optional[int] = None, item_type : typing.Optional[typing.Tuple] = None, 
            accept_list : bool = False, accept_item : bool = False, deepcopy_default : bool = False, 
            allow_None : bool = False, doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False,
            per_instance_descriptor : bool = False, class_member : bool = False, 
            fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, bounds=bounds, length=length, item_type=item_type, doc=doc, constant=constant, 
            readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)
        self.accept_list = accept_list
        self.accept_item = accept_item
        self.dtype = (tuple,) # re-assigned
      
    def validate_and_adapt(self, value: typing.Any) -> typing.Tuple:
        if self.accept_list and isinstance(value, list):
            value = tuple(value) 
        if self.accept_item and not isinstance(value, (list, tuple, type(None))):
            value = (value,)
        self._assert(value, self.bounds, self.length, self.dtype, self.item_type, self.allow_None)
        return value 
    
    @classmethod
    def serialize(cls, value):
        if value is None:
            return None
        return list(value) # As JSON has no tuple representation

    @classmethod
    def deserialize(cls, value):
        if value == 'null':
            return None
        return tuple(value) # As JSON has no tuple representation



class List(Iterable):
    """
    Parameter whose value is a list of objects, usually of a specified type.

    The bounds allow a minimum and/or maximum length of
    list to be enforced.  If the item_type is non-None, all
    items in the list are checked to be of that type.

    `class_` is accepted as an alias for `item_type`, but is
    deprecated due to conflict with how the `class_` slot is
    used in Selector classes.
    """

    __slots__ = ['accept_tuple']

    def __init__(self, default: typing.Any, *, bounds : typing.Optional[typing.Tuple[int, int]] = None, 
            length : typing.Optional[int] = None, item_type : typing.Optional[typing.Tuple] = None, 
            accept_tuple : bool = False, deepcopy_default : bool = False, 
            allow_None : bool = False, doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, 
            per_instance_descriptor : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None,
            fset : typing.Optional[typing.Callable] = None, fdel : typing.Optional[typing.Callable] = None,
            precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, bounds=bounds, length=length, item_type=item_type,
            doc=doc, constant=constant, readonly=readonly, allow_None=allow_None, 
            per_instance_descriptor=per_instance_descriptor, deepcopy_default=deepcopy_default, 
            class_member=class_member, fget=fget, fset=fset, fdel=fdel, precedence=precedence)
        self.accept_tuple = accept_tuple
        self.dtype = list

    def validate_and_adapt(self, value: typing.Any) -> typing.Tuple:
        if self.accept_tuple and isinstance(value, tuple):
            value = list(value)
        self._assert(value, self.bounds, self.length, self.dtype, self.item_type, self.allow_None)
        return value 



class Callable(Parameter):
    """
    Parameter holding a value that is a callable object, such as a function.

    A keyword argument instantiate=True should be provided when a
    function object is used that might have state.  On the other hand,
    regular standalone functions cannot be deepcopied as of Python
    2.4, so instantiate must be False for those values.
    """

    def validate_and_adapt(self, value : typing.Any) -> typing.Callable:
        if (self.allow_None and value is None) or callable(value):
            return value
        raise_ValueError("given value not a callable object, but type {}.".format(type(value)), self)
       


class Composite(Parameter):
    """
    A Parameter that is a composite of a set of other attributes of the class.

    The constructor argument 'attribs' takes a list of attribute
    names, which may or may not be Parameters.  Getting the parameter
    returns a list of the values of the constituents of the composite,
    in the order specified.  Likewise, setting the parameter takes a
    sequence of values and sets the value of the constituent
    attributes.

    This Parameter type has not been tested with watchers and
    dependencies, and may not support them properly.
    """

    __slots__ = ['attribs']

    def __init__(self, attribs : typing.List[typing.Union[str, Parameter]], *, 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(None, doc=doc, constant=constant, readonly=readonly, allow_None=allow_None, 
            per_instance_descriptor=per_instance_descriptor, deepcopy_default=deepcopy_default, 
            class_member=class_member, fget=fget, fset=fset, fdel=fdel, precedence=precedence)
        self.attribs = [] 
        if attribs is not None:
            for attrib in attribs:
                if isinstance(attrib, Parameter):
                    self.attribs.append(attrib.name)
                else:
                    self.attribs.append(attrib)            
        
    def __get__(self, obj, objtype) -> typing.List[typing.Any]:
        """
        Return the values of all the attribs, as a list.
        """
        return [getattr(obj, attr) for attr in self.attribs]

    def validate_and_adapt(self, value):
        if not len(value) == len(self.attribs):
            raise_ValueError("Compound parameter got the wrong number of values (needed {}, but got {}).".format(
                        len(self.attribs), len(value)), self)
        return value

    def _post_setter(self, obj, val):
        for a, v in zip(self.attribs, val):
            setattr(obj, a, v)



class SelectorBase(Parameter):
    """
    Parameter whose value must be chosen from a list of possibilities.

    Subclasses must implement get_range().
    """

    __abstract = True

    @property
    def range(self):
        raise NotImplementedError("get_range() must be implemented in subclasses.")



class Selector(SelectorBase):
    """
    Parameter whose value must be one object from a list of possible objects.

    By default, if no default is specified, picks the first object from
    the provided set of objects, as long as the objects are in an
    ordered data collection.

    check_on_set restricts the value to be among the current list of
    objects. By default, if objects are initially supplied,
    check_on_set is True, whereas if no objects are initially
    supplied, check_on_set is False. This can be overridden by
    explicitly specifying check_on_set initially.

    If check_on_set is True (either because objects are supplied
    initially, or because it is explicitly specified), the default
    (initial) value must be among the list of objects (unless the
    default value is None).

    The list of objects can be supplied as a list (appropriate for
    selecting among a set of strings, or among a set of objects with a
    "name" parameter), or as a (preferably ordered) dictionary from
    names to objects.  If a dictionary is supplied, the objects
    will need to be hashable so that their names can be looked
    up from the object value.
    """

    __slots__ = ['objects', 'names']

    # Selector is usually used to allow selection from a list of
    # existing objects, therefore instantiate is False by default.
    def __init__(self, *, objects : typing.List[typing.Any], default : typing.Any, empty_default : bool = False,  
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, 
            fset : typing.Optional[typing.Callable] = None, fdel : typing.Optional[typing.Callable] = None, 
            precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, doc=doc, constant=constant, readonly=readonly, 
            allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, deepcopy_default=deepcopy_default, 
            class_member=class_member, fget=fget, fset=fset, fdel=fdel, precedence=precedence)
        if objects is None:
            objects = []
            autodefault = None
        elif isinstance(objects, collections.abc.Mapping):
            self.names = objects
            self.objects = list(objects.values())
            autodefault = self.objects[0]
        elif isinstance(objects, (list, tuple)):
            self.names = None
            self.objects = objects
            autodefault = objects[0]
        else:
            raise TypeError("objects should be a list, tuple, mapping or None. Given type : {}".format(type(objects)))
        default = autodefault if (not empty_default and default is None) else default

    def validate_and_adapt(self, value: typing.Any) -> typing.Any:
        """
        val must be None or one of the objects in self.objects.
        """
        if not (value in self.objects or (self.allow_None and value is None)):
            raise_ValueError("given value not in list of possible objects, valid options include {}".format(
                get_iterable_printfriendly_repr(self.objects)), self)
        return value

    @property
    def range(self):
        """
        Return the possible objects to which this parameter could be set.

        (Returns the dictionary {object.name:object}.)
        """
        if self.names is not None:
            return named_objs(self.objects, self.names)
        else:
            return self.objects



class ClassSelector(SelectorBase):
    """
    Parameter allowing selection of either a subclass or an instance of a given set of classes.
    By default, requires an instance, but if isinstance=False, accepts a class instead.
    Both class and instance values respect the instantiate slot, though it matters only
    for isinstance=True.
    """

    __slots__ = ['class_', 'isinstance']

    def __init__(self, *, class_ , default : typing.Any, isinstance : bool = True, deepcopy_default : bool = False,  
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, class_member : bool = False, 
            fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, doc=doc, constant=constant, readonly=readonly, 
            allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, deepcopy_default=deepcopy_default, 
            class_member=class_member, fget=fget, fset=fset, fdel=fdel, precedence=precedence)
        self.class_ = class_
        self.isinstance = isinstance

    def _get_class_name(self):
        if isinstance(self.class_, tuple):
            return ('(%s)' % ', '.join(cl.__name__ for cl in self.class_))
        else:
            return self.class_.__name__

    def validate_and_adapt(self, value): 
        if (value is None and self.allow_None):
            return
        if self.isinstance:
            if not isinstance(value, self.class_):
                raise_ValueError("{} parameter {} value must be an instance of {}, not {}.".format(
                    self.__class__.__name__, self.name, self._get_class_name(), value), self)
        else:
            try:
                if not issubclass(value, self.class_):
                    raise_ValueError("{} parameter {} must be a subclass of {}, not {}.".format(
                        self.__class__.__name__, self.name, self._get_class_name(), value.__name__), self)
            except TypeError as ex:
                if str(ex).startswith("ssubclass() arg 1 must be a class"):
                    raise_ValueError("Value must be a class, not an instance.", self) 
                raise ex from None # raise other type errors anyway                  
        return value

    @property
    def range(self):
        """
        Return the possible types for this parameter's value.

        (I.e. return `{name: <class>}` for all classes that are
        concrete_descendents() of `self.class_`.)

        Only classes from modules that have been imported are added
        (see concrete_descendents()).
        """
        classes = self.class_ if isinstance(self.class_, tuple) else (self.class_,)
        all_classes = {}
        for cls in classes:
            all_classes.update(concrete_descendents(cls))
        d = OrderedDict((name, class_) for name,class_ in all_classes.items())
        if self.allow_None:
            d['None'] = None
        return d



class TupleSelector(Selector):
    """
    Variant of Selector where the value can be multiple objects from
    a list of possible objects.
    """

    # Changed from ListSelector. Iterables need to be frozen to prevent spurious addition. 
    # To prevent duplicates, use frozen set selector

    __slots__ = ['accept_list']

    def __init__(self, *, objects : typing.List, default : typing.Any, accept_list : bool = True,
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(objects=objects, default=default, empty_default=True, doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)
        self.accept_list = accept_list

    def validate_and_adapt(self, value : typing.Any):
        if value is None and self.allow_None:
            return
        if value not in self.objects: 
            # i.e. without iterating, we check that the value is not present in the objects
            # This is useful to have list or iterables themselves as part of objects
            # let objects = [[1,2], 3 ,4], if [1,2] is passed, then we should try to accept it plainly before moving to iterating
            # and checking
            if isinstance(value, list) and self.accept_list:
                value = tuple(value)
            if not isinstance(value, tuple):
                raise_ValueError(f"object {value} not specified as a valid member of list of objects.", self)
            else:
                for obj in value:
                    if obj not in self.objects: 
                        raise_ValueError("object {} not specified as a valid member of list of objects.".format(obj), self)
        return value


# For portable code:
#   - specify paths in unix (rather than Windows) style;
#   - use resolve_path(path_to_file=True) for paths to existing files to be read,
#   - use resolve_path(path_to_file=False) for paths to existing folders to be read,
#     and normalize_path() for paths to new files to be written.

class resolve_path(ParameterizedFunction):
    """
    Find the path to an existing file, searching the paths specified
    in the search_paths parameter if the filename is not absolute, and
    converting a UNIX-style path to the current OS's format if
    necessary.

    To turn a supplied relative path into an absolute one, the path is
    appended to paths in the search_paths parameter, in order, until
    the file is found.

    An IOError is raised if the file is not found.

    Similar to Python's os.path.abspath(), except more search paths
    than just os.getcwd() can be used, and the file must exist.
    """

    search_paths = List(default=[os.getcwd()], doc="""
        Prepended to a non-relative path, in order, until a file is
        found.""")

    path_to_file = Boolean(default=True, 
                           allow_None=True, doc="""
        String specifying whether the path refers to a 'File' or a
        'Folder'. If None, the path may point to *either* a 'File' *or*
        a 'Folder'.""")

    def __call__(self, path : str, **params) -> str:
        p = ParamOverrides(self, params)
        path = os.path.normpath(path)
        ftype = "File" if p.path_to_file is True \
            else "Folder" if p.path_to_file is False else "Path"

        if not p.search_paths:
            p.search_paths = [os.getcwd()]

        if os.path.isabs(path):
            if ((p.path_to_file is None  and os.path.exists(path)) or
                (p.path_to_file is True  and os.path.isfile(path)) or
                (p.path_to_file is False and os.path.isdir( path))):
                return path
            raise IOError("%s '%s' not found." % (ftype,path))

        else:
            paths_tried = []
            for prefix in p.search_paths:
                try_path = os.path.join(os.path.normpath(prefix), path)

                if ((p.path_to_file is None  and os.path.exists(try_path)) or
                    (p.path_to_file is True  and os.path.isfile(try_path)) or
                    (p.path_to_file is False and os.path.isdir( try_path))):
                    return try_path

                paths_tried.append(try_path)

            raise IOError(ftype + " " + os.path.split(path)[1] + " was not found in the following place(s): " + str(paths_tried) + ".")


class normalize_path(ParameterizedFunction):
    """
    Convert a UNIX-style path to the current OS's format,
    typically for creating a new file or directory.

    If the path is not already absolute, it will be made absolute
    (using the prefix parameter).

    Should do the same as Python's os.path.abspath(), except using
    prefix rather than os.getcwd).
    """

    prefix = String(default=os.getcwd(),doc="""
        Prepended to the specified path, if that path is not
        absolute.""")

    def __call__(self, path : str = "", **params):
        p = ParamOverrides(self,params)

        if not os.path.isabs(path):
            path = os.path.join(os.path.normpath(p.prefix), path)

        return os.path.normpath(path)



class Path(Parameter):
    """
    Parameter that can be set to a string specifying the path of a file or folder.

    The string should be specified in UNIX style, but it will be
    returned in the format of the user's operating system. Please use
    the Filename or Foldername classes if you require discrimination
    between the two possibilities.

    The specified path can be absolute, or relative to either:

    * any of the paths specified in the search_paths attribute (if
       search_paths is not None);

    or

    * any of the paths searched by resolve_path() (if search_paths
      is None).
    """

    __slots__ = ['search_paths']

    def __init__(self, default : typing.Any = '', *, search_paths : typing.Optional[str] = None, 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False, 
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)
        if isinstance(search_paths, str):
            self.search_paths = [search_paths]
        elif isinstance(search_paths, list):
            self.search_paths = search_paths
        else: 
            self.search_paths = []

    def _resolve(self, path):
        return resolve_path(path, path_to_file=None, search_paths=self.search_paths)

    def validate_and_adapt(self, value : typing.Any) -> typing.Any:
        if value is None and self.allow_None:
            return 
        else:
            return self._resolve(value)
         
    def __get__(self, obj, objtype) -> str:
        """
        Return an absolute, normalized path (see resolve_path).
        """
        raw_path = super().__get__(obj, objtype)
        return None if raw_path is None else self._resolve(raw_path)

    def __getstate__(self):
        # don't want to pickle the search_paths
        state = super().__getstate__()
        if 'search_paths' in state:
            state['search_paths'] = []
        return state



class Filename(Path):
    """
    Parameter that can be set to a string specifying the path of a file.

    The string should be specified in UNIX style, but it will be
    returned in the format of the user's operating system.

    The specified path can be absolute, or relative to either:

    * any of the paths specified in the search_paths attribute (if
      search_paths is not None);

    or

    * any of the paths searched by resolve_path() (if search_paths
      is None).
    """

    def _resolve(self, path):
        return resolve_path(path, path_to_file=True, search_paths=self.search_paths)


class Foldername(Path):
    """
    Parameter that can be set to a string specifying the path of a folder.

    The string should be specified in UNIX style, but it will be
    returned in the format of the user's operating system.

    The specified path can be absolute, or relative to either:

    * any of the paths specified in the search_paths attribute (if
      search_paths is not None);

    or

    * any of the paths searched by resolve_dir_path() (if search_paths
      is None).
    """

    def _resolve(self, path):
        return resolve_path(path, path_to_file=False, search_paths=self.search_paths)



def abbreviate_paths(pathspec,named_paths):
    """
    Given a dict of (pathname,path) pairs, removes any prefix shared by all pathnames.
    Helps keep menu items short yet unambiguous.
    """

    prefix = os.path.commonprefix([os.path.dirname(name)+os.path.sep for name in named_paths.keys()]+[pathspec])
    return OrderedDict([(name[len(prefix):],path) for name,path in named_paths.items()])



class FileSelector(Selector):
    """
    Given a path glob, allows one file to be selected from those matching.
    """
    __slots__ = ['path']

    def __init__(self, default : typing.Any, *, objects : typing.List, path : str = "", 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False, 
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, objects=objects, empty_default=True, doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)
        self.path = path # update is automatically called

    def _post_slot_set(self, slot: str, old : typing.Any, value : typing.Any) -> None:
        super()._post_slot_set(slot, old, value)
        if slot == 'path':
            self.update()

    def update(self):
        self.objects = sorted(glob.glob(self.path))
        if self.default in self.objects:
            return
        self.default = self.objects[0] if self.objects else None

    @property
    def range(self):
        return abbreviate_paths(self.path, super().range)



class MultiFileSelector(FileSelector):
    """
    Given a path glob, allows multiple files to be selected from the list of matches.
    """
    __slots__ = ['path']

    def __init__(self, default : typing.Any, *, path : str = "", 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False, 
            label : typing.Optional[str] = None, per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, objects=None, doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, 
            fdel=fdel, precedence=precedence)
       
    def update(self):
        self.objects = sorted(glob.glob(self.path))
        if self.default and all([o in self.objects for o in self.default]):
            return
        self.default = self.objects



class Date(Number):
    """
    Date parameter of datetime or date type.
    """

    def __init__(self, default, *, bounds : typing.Union[typing.Tuple, None] = None, 
            crop_to_bounds : bool = False, inclusive_bounds : typing.Tuple = (True,True), step : typing.Any = None, 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, bounds=bounds, crop_to_bounds=crop_to_bounds,
            inclusive_bounds=inclusive_bounds, step=step, doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)
        self.dtype = dt_types

    def _validate_step(self, val):
        if self.step is not None and not isinstance(self.step, dt_types):
            raise ValueError(f"Step can only be None, a datetime or datetime type, not type {type(val)}")

    @classmethod
    def serialize(cls, value):
        if value is None:
            return None
        if not isinstance(value, (dt.datetime, dt.date)): # i.e np.datetime64, note numpy is imported only on requirement
            value = value.astype(dt.datetime)
        return value.strftime("%Y-%m-%dT%H:%M:%S.%f")

    @classmethod
    def deserialize(cls, value):
        if value == None:
            return None
        return dt.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")



class CalendarDate(Number):
    """
    Parameter specifically allowing dates (not datetimes).
    """

    def __init__(self, default, *, bounds : typing.Union[typing.Tuple, None] = None, 
            crop_to_bounds : bool = False, inclusive_bounds : typing.Tuple = (True,True), step : typing.Any = None, 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, bounds=bounds, crop_to_bounds=crop_to_bounds,
            inclusive_bounds=inclusive_bounds, step=step, doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)
        self.dtype = dt.date

    def _validate_step(self, step):
        if step is not None and not isinstance(step, self.dtype):
            raise ValueError("Step can only be None or a date type.")

    @classmethod
    def serialize(cls, value):
        if value is None:
            return None
        return value.strftime("%Y-%m-%d")

    @classmethod
    def deserialize(cls, value):
        if value is None:
            return None
        return dt.datetime.strptime(value, "%Y-%m-%d").date()



class CSS3Color(Parameter):
    """
    Color parameter defined as a hex RGB string with an optional #
    prefix or (optionally) as a CSS3 color name.
    """

    # CSS3 color specification https://www.w3.org/TR/css-color-3/#svg-color
    _named_colors = [ 'aliceblue', 'antiquewhite', 'aqua',
        'aquamarine', 'azure', 'beige', 'bisque', 'black',
        'blanchedalmond', 'blue', 'blueviolet', 'brown', 'burlywood',
        'cadetblue', 'chartreuse', 'chocolate', 'coral',
        'cornflowerblue', 'cornsilk', 'crimson', 'cyan', 'darkblue',
        'darkcyan', 'darkgoldenrod', 'darkgray', 'darkgrey',
        'darkgreen', 'darkkhaki', 'darkmagenta', 'darkolivegreen',
        'darkorange', 'darkorchid', 'darkred', 'darksalmon',
        'darkseagreen', 'darkslateblue', 'darkslategray',
        'darkslategrey', 'darkturquoise', 'darkviolet', 'deeppink',
        'deepskyblue', 'dimgray', 'dimgrey', 'dodgerblue',
        'firebrick', 'floralwhite', 'forestgreen', 'fuchsia',
        'gainsboro', 'ghostwhite', 'gold', 'goldenrod', 'gray',
        'grey', 'green', 'greenyellow', 'honeydew', 'hotpink',
        'indianred', 'indigo', 'ivory', 'khaki', 'lavender',
        'lavenderblush', 'lawngreen', 'lemonchiffon', 'lightblue',
        'lightcoral', 'lightcyan', 'lightgoldenrodyellow',
        'lightgray', 'lightgrey', 'lightgreen', 'lightpink',
        'lightsalmon', 'lightseagreen', 'lightskyblue',
        'lightslategray', 'lightslategrey', 'lightsteelblue',
        'lightyellow', 'lime', 'limegreen', 'linen', 'magenta',
        'maroon', 'mediumaquamarine', 'mediumblue', 'mediumorchid',
        'mediumpurple', 'mediumseagreen', 'mediumslateblue',
        'mediumspringgreen', 'mediumturquoise', 'mediumvioletred',
        'midnightblue', 'mintcream', 'mistyrose', 'moccasin',
        'navajowhite', 'navy', 'oldlace', 'olive', 'olivedrab',
        'orange', 'orangered', 'orchid', 'palegoldenrod', 'palegreen',
        'paleturquoise', 'palevioletred', 'papayawhip', 'peachpuff',
        'peru', 'pink', 'plum', 'powderblue', 'purple', 'red',
        'rosybrown', 'royalblue', 'saddlebrown', 'salmon',
        'sandybrown', 'seagreen', 'seashell', 'sienna', 'silver',
        'skyblue', 'slateblue', 'slategray', 'slategrey', 'snow',
        'springgreen', 'steelblue', 'tan', 'teal', 'thistle',
        'tomato', 'turquoise', 'violet', 'wheat', 'white',
        'whitesmoke', 'yellow', 'yellowgreen']

    __slots__ = ['allow_named']

    def __init__(self, default, *, allow_named : bool = True,  doc : typing.Optional[str] = None, constant : bool = False, 
            readonly : bool = False, allow_None : bool = False,
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        super().__init__(default=default, doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel,  
            precedence=precedence)
        self.allow_named = allow_named
    
    def validate_and_adapt(self, value : typing.Any):
        if (self.allow_None and value is None):
            return
        if not isinstance(value, str):
            raise ValueError("Color parameter %r expects a string value, "
                             "not an object of type %s." % (self.name, type(value)))
        if self.allow_named and value in self._named_colors:
            return 
        is_hex = re.match('^#?(([0-9a-fA-F]{2}){3}|([0-9a-fA-F]){3})$', value)
        if not is_hex:
            raise ValueError("Color '%s' only takes RGB hex codes "
                                 "or named colors, received '%s'." % (self.name, value))
      


class Range(Tuple):
    """
    A numeric range with optional bounds and softbounds.
    """

    __slots__ = ['bounds', 'inclusive_bounds', 'softbounds', 'step']

    def __init__(self, default : typing.Optional[typing.Tuple] = None, *, bounds: typing.Optional[typing.Tuple[int, int]] = None, 
            length : typing.Optional[int] = None, item_type : typing.Optional[typing.Tuple] = None, 
            softbounds=None, inclusive_bounds=(True,True), step=None, 
            doc : typing.Optional[str] = None, constant : bool = False, 
            readonly : bool = False, allow_None : bool = False, label : typing.Optional[str] = None,  
            per_instance_descriptor : bool = False, deepcopy_default : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        self.inclusive_bounds = inclusive_bounds
        self.softbounds = softbounds
        self.step = step
        super().__init__(default=default, bounds=bounds, item_type=item_type, length=length, doc=doc, 
            constant=constant, readonly=readonly, allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, 
            deepcopy_default=deepcopy_default, class_member=class_member, fget=fget, fset=fset, fdel=fdel, 
            precedence=precedence)

    def validate_and_adapt(self, value : typing.Any) -> typing.Tuple:
        raise NotImplementedError("Range validation not implemented")
        super()._validate(val)
        self._validate_bounds(val, self.bounds, self.inclusive_bounds)

    def _validate_bounds(self, val, bounds, inclusive_bounds):
        if bounds is None or (val is None and self.allow_None):
            return
        vmin, vmax = bounds
        incmin, incmax = inclusive_bounds
        for bound, v in zip(['lower', 'upper'], val):
            too_low = (vmin is not None) and (v < vmin if incmin else v <= vmin)
            too_high = (vmax is not None) and (v > vmax if incmax else v >= vmax)
            if too_low or too_high:
                raise ValueError("Range parameter %r's %s bound must be in range %s."
                                 % (self.name, bound, self.rangestr()))

    @property
    def rangestr(self):
        vmin, vmax = self.bounds
        incmin, incmax = self.inclusive_bounds
        incmin = '[' if incmin else '('
        incmax = ']' if incmax else ')'
        return '%s%s, %s%s' % (incmin, vmin, vmax, incmax)



class DateRange(Range):
    """
    A datetime or date range specified as (start, end).

    Bounds must be specified as datetime or date types (see param.dt_types).
    """

    def _validate_value(self, val, allow_None):
        # Cannot use super()._validate_value as DateRange inherits from
        # NumericTuple which check that the tuple values are numbers and
        # datetime objects aren't numbers.
        if allow_None and val is None:
            return

        if not isinstance(val, tuple):
            raise ValueError("DateRange parameter %r only takes a tuple value, "
                             "not %s." % (self.name, type(val).__name__))
        for n in val:
            if isinstance(n, dt_types):
                continue
            raise ValueError("DateRange parameter %r only takes date/datetime "
                             "values, not type %s." % (self.name, type(n).__name__))

        start, end = val
        if not end >= start:
            raise ValueError("DateRange parameter %r's end datetime %s "
                             "is before start datetime %s." %
                             (self.name, val[1], val[0]))

    @classmethod
    def serialize(cls, value):
        if value is None:
            return 'null'
        # List as JSON has no tuple representation
        serialized = []
        for v in value:
            if not isinstance(v, (dt.datetime, dt.date)): # i.e np.datetime64
                v = v.astype(dt.datetime)
            # Separate date and datetime to deserialize to the right type.
            if type(v) == dt.date:
                v = v.strftime("%Y-%m-%d")
            else:
                v = v.strftime("%Y-%m-%dT%H:%M:%S.%f")
            serialized.append(v)
        return serialized

    def deserialize(cls, value):
        if value == 'null':
            return None
        deserialized = []
        for v in value:
            # Date
            if len(v) == 10:
                v = dt.datetime.strptime(v, "%Y-%m-%d").date()
            # Datetime
            else:
                v = dt.datetime.strptime(v, "%Y-%m-%dT%H:%M:%S.%f")
            deserialized.append(v)
        # As JSON has no tuple representation
        return tuple(deserialized)



class CalendarDateRange(Range):
    """
    A date range specified as (start_date, end_date).
    """
    def _validate_value(self, val, allow_None):
        if allow_None and val is None:
            return

        for n in val:
            if not isinstance(n, dt.date):
                raise ValueError("CalendarDateRange parameter %r only "
                                 "takes date types, not %s." % (self.name, val))

        start, end = val
        if not end >= start:
            raise ValueError("CalendarDateRange parameter %r's end date "
                             "%s is before start date %s." %
                             (self.name, val[1], val[0]))

    @classmethod
    def serialize(cls, value):
        if value is None:
            return 'null'
        # As JSON has no tuple representation
        return [v.strftime("%Y-%m-%d") for v in value]

    @classmethod
    def deserialize(cls, value):
        if value == 'null':
            return None
        # As JSON has no tuple representation
        return tuple([dt.datetime.strptime(v, "%Y-%m-%d").date() for v in value])
    



def get_typed_iterable_bounds(bounds : tuple) -> tuple:
    if bounds[0] is None and bounds[1] is None:
        bounds = (0, 2*sys.maxsize + 1)
    elif bounds[0] is None: 
        bounds = (0, bounds[1])
    elif bounds[1] is None:
        bounds = (bounds[0], 2*sys.maxsize + 1)
    return bounds


class BaseConstrainedList(collections.abc.MutableSequence):

    # Need to check mul
    
    def __init__(self, default : typing.List[typing.Any], *, bounds : tuple = (0, None), 
                constant : bool = False, skip_validate : bool = False) -> None:
        super().__init__()
        self.constant = constant
        self.bounds   = get_typed_iterable_bounds(bounds)
        if not skip_validate:
            self._validate_for_set(default)    
        self._inner = default
  
    def _validate_for_set(self, value : typing.Any) -> None:
        self._validate_value(value)
        self._validate_bounds_for_set(value)
        self._validate_items(value)

    def _validate_for_extension(self, value : typing.List) -> None:
        if self.constant:
            raise ValueError(f"List {get_iterable_printfriendly_repr(self._inner)} is a constant, cannot be modified.")
        self._validate_value(value)
        self._validate_bounds_for_extension(value)
        self._validate_items(value)

    def _validate_for_insertion(self, value : typing.Any) -> None:
        if self.constant:
            raise ValueError(f"List {get_iterable_printfriendly_repr(self._inner)} is a constant, cannot be modified.")
        self._validate_bounds_for_extension()
        self._validate_item(value)

    def _validate_value(self, value : typing.Any) -> None:
        if not isinstance(value, list):
            raise TypeError(f"Given value for a constrained list is not a list, but type {type(value)}") 
           
    def _validate_items(self, value : typing.Any) -> None:
        raise NotImplementedError("Please implement _validate_item in the child of BaseConstrainedList.")

    def _validate_item(self, value : typing.Any):
        raise NotImplementedError("Please implement _validate_single_item in the child of BaseConstrainedList.")

    def _validate_bounds_for_set(self, value : typing.Any) -> None:
        if not (value.__len__() >= self.bounds[0] and value.__len__() <= self.bounds[1]):
            raise ValueError(wrap_error_text(
                f"""given list {get_iterable_printfriendly_repr(value)} has length out of bounds {self.bounds}. 
                given length : {value.__len__()}"""))
               
    def _validate_bounds_for_extension(self,  value : typing.Any = [None]) -> None:
        if not (self._inner.__len__() + value.__len__() >= self.bounds[0] and 
                            self._inner.__len__() + value.__len__() <= self.bounds[1]):
            raise ValueError(wrap_error_text(
                f"""given list for extending {get_iterable_printfriendly_repr(value)} extends existing list longer 
                than bounds {self.bounds}. given length : {self._inner.__len__() + value.__len__()}"""))         

    def __len__(self) -> int:
        return self._inner.__len__()

    def __iter__(self) -> typing.Any:
        return self._inner.__iter__()

    def __str__(self) -> str:
        return self._inner.__str__()

    def __contains__(self, item : typing.Any) -> bool:
        return item in self._inner

    def __getitem__(self, index : int):
        return self._inner[index]

    def __setitem__(self, index : int, value : typing.Any) -> None:
        if self.constant:
            raise ValueError(f"List {get_iterable_printfriendly_repr(self._inner)} is a constant, cannot be modified.")
        self._validate_item(value)
        self._inner[index] = value

    def __delitem__(self, index : int) -> None:
        del self._inner[index]

    def __repr__(self) -> str:
        return self._inner.__repr__()
    
    def __imul__(self, value : typing.Any) -> typing.List:
        return self._inner.__imul__(value)

    def __mul__(self, value : typing.Any) -> typing.List:
        return self._inner.__mul__(value)

    def __sizeof__(self) -> int:
        return self._inner.__sizeof__()

    def __lt__(self, __x : typing.List[typing.Any]) -> bool:
        return self._inner.__lt__(__x)
    
    def __le__(self, __x : typing.List[typing.Any]) -> bool:
        return self._inner.__le__(__x)
    
    def __eq__(self, __x : typing.List[typing.Any]) -> bool:  
        return self._inner.__eq__(__x)
    
    def __ne__(self, __x : typing.List[typing.Any]) -> bool:
        return self._inner.__ne__(__x)
    
    def __gt__(self, __x : typing.List[typing.Any]) -> bool:
        return  self._inner.__gt__(__x)
    
    def __ge__(self, __x : typing.List[typing.Any]) -> bool:
        return self._inner.__ge__(__x)
       
    def __rmul__(self, __n : int) -> typing.List:
        return self._inner.__rmul__(__n)

    def __reversed__(self) -> typing.Iterator:
        return self._inner.__reversed__()

    def __add__(self, __x : typing.List[typing.Any]) -> typing.List: 
        if isinstance(__x, self.__class__):
            return self._inner.__add__(__x._inner)
        else:
            return self._inner.__add__(__x)

    def __iadd__(self, values : typing.List[typing.Any] ) -> typing.List:
        raise NotImplementedError("Please implement __iadd__ in the child of BaseConstrainedList.")

    def insert(self, __index : int, __object : typing.Any) -> None:
        self._validate_for_insertion(__object)
        self._inner.insert(__index, __object)

    def append(self, __object : typing.Any) -> None:
        self._validate_for_insertion(__object)
        self._inner.append(__object)
        
    def extend(self, __iterable) -> None:
        self._validate_for_extension(__iterable)
        self._inner.extend(__iterable)

    def reverse(self) -> None:
        self._inner.reverse()
    
    def pop(self, __index: int) -> typing.Any:
        return self._inner.pop(__index)

    def count(self, __value : typing.Any) -> int:
        return self._inner.count(__value)

    def clear(self) -> None: 
        self._inner.clear()

    def index(self, __value : typing.Any, __start : int, __stop : int) -> int:
        return self._inner.index(__value, __start, __stop)

    def remove(self, __value : typing.Any) -> None:
        self._inner.remove(__value)

    def sort(self, key : typing.Any, reverse : bool):
        self._inner.sort(key=key, reverse=reverse)

    def copy(self, return_as_typed_list : bool = False):
        raise NotImplementedError("Please implement copy() in the child of BaseConstrainedList.")


       
class TypeConstrainedList(BaseConstrainedList):
 
    def __init__(self, default : typing.List, *, item_type : typing.Any = None, 
                bounds : tuple = (0,None), constant : bool = False, skip_validate : bool = False) -> None:
        self.item_type = item_type
        super().__init__(default=default, bounds=bounds, constant=constant, skip_validate=skip_validate)
        
    def _validate_items(self, value : typing.Any) -> None:
        if self.item_type is not None:
            for val in value:  
                if not isinstance(val, self.item_type):
                    raise TypeError(
                        wrap_error_text(f"""
                        Not all elements of list {get_iterable_printfriendly_repr(value)} given are of allowed item type(s), 
                        which are : {self.item_type}. Given type {type(val)}. Cannot set or extend typed list."""
                    ))
                
    def _validate_item(self, value : typing.Any):
        if self.item_type is not None and not isinstance(value, self.item_type):
            raise TypeError(
                wrap_error_text(f"""
                    Not all elements given are of allowed item type(s), which are : {self.item_type}. 
                    Given type {type(value)}. Cannot append or insert in typed list."""
            ))
      
    def __iadd__(self, value : typing.List[typing.Any]):
        self._validate_for_extension(value)
        return TypeConstrainedList(default=self._inner.__iadd__(value), item_type=self.item_type, bounds=self.bounds,
                                constant=self.constant, skip_validate=True)
    
    def copy(self, return_as_typed_list : bool = False) -> typing.Union["TypeConstrainedList", typing.List[typing.Any]]: 
        if return_as_typed_list:
            return TypeConstrainedList(default=self._inner.copy(), item_type=self.item_type, bounds=self.bounds, 
                            constant=self.constant, skip_validate=True)
        else:
            return self._inner.copy()
      


class TypedList(ClassSelector):
    
    __slots__ = ['item_type', 'bounds', 'accept_nonlist_object']

    def __init__(self, default : typing.Optional[typing.List[typing.Any]] = None, *, item_type : typing.Any = None, 
            deepcopy_default : bool = True, accept_nonlist_object : bool = False, 
            allow_None : bool = True,  bounds : tuple = (0,None), 
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, 
            per_instance_descriptor : bool = False, class_member : bool = False, 
            fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        if default is not None:
            default = TypeConstrainedList(default=default, item_type=item_type, bounds=bounds, constant=constant, 
                               skip_validate=False) # type: ignore    
        super().__init__(class_ = TypeConstrainedList, default=default, isinstance=True, deepcopy_default=deepcopy_default, 
            doc=doc, constant=constant, readonly=readonly, allow_None=allow_None, 
            per_instance_descriptor=per_instance_descriptor, class_member=class_member, fget=fget, fset=fset, 
            fdel=fdel, precedence=precedence)
        self.item_type = item_type
        self.bounds    = bounds 
        self.accept_nonlist_object = accept_nonlist_object

    # @instance_descriptor - super().__set__ takes care of instance descriptors 
    def validate_and_adapt(self, value : typing.Any):
        if self.allow_None and value is None:
            return
        if value is not None and self.accept_nonlist_object and not isinstance(value, list):
            value = [value]   
        return TypeConstrainedList(default=value, item_type=self.item_type, bounds=self.bounds, 
                        constant=self.constant, skip_validate=False)        
            
    @classmethod
    def serialize(cls, value : TypeConstrainedList) -> typing.Any:
        if value is None:
            return None
        return value._inner

    # no need for deserialize, when __set__ is called TypeConstrainedList is automatically created  
    

class TypeConstrainedDict(collections.abc.MutableMapping):
    """ A dictionary which contains only ``NewDict`` values. """

    def __init__(self, default : typing.Dict, *, key_type : typing.Optional[typing.Union[type, typing.Tuple]] = None, 
                item_type : typing.Optional[typing.Union[type, typing.Tuple]] = None, 
                bounds : typing.Tuple = (0, None), constant : bool = False, skip_validate : bool = False):
        super().__init__()
        self.key_type  = key_type
        self.item_type = item_type
        self.bounds    = get_typed_iterable_bounds(bounds)
        self.constant  = constant 
        if not skip_validate:
            self._validate_for_set(default)
        self._inner = default

    def _validate_for_set(self, value : typing.Dict) -> None:
        self._validate_value(value)
        self._validate_bounds_for_set(value)
        self._validate_items(value)

    def _validate_for_insertion(self, value : typing.Dict) -> None:
        if self.constant:
            raise ValueError(f"Dict {get_iterable_printfriendly_repr(self._inner)} is a constant and cannot be modified.")
        self._validate_value(value)
        self._validate_bounds_for_extension(value)
        self._validate_items(value)

    def _validate_value(self, value) -> None:
        if not isinstance(value, dict):
            raise TypeError(wrap_error_text(f"""
                Given value for typed dictionary is not a dictionary. Given type : {type(value)}. Expected dictionary."""))

    def _validate_bounds_for_set(self, value : typing.Dict) -> None:
        if not (self.bounds[0] <= value.__len__() <= self.bounds[1]):
            raise ValueError(wrap_error_text(f"""
                Given dictionary length outside bounds. Given length {value.__len__()}, expected length : {self.bounds}"""))
                
    def _validate_bounds_for_extension(self, value : typing.Dict = {"dummy" : "dummy"}) -> None:
        if not (self.bounds[0] <= self._inner.__len__() + value.__len__() <= self.bounds[1]):
            raise ValueError(wrap_error_text(f"""
                Extending dictionary crosses bounds. Existing length {self._inner.__len__()}, 
                length of items to be added : {value.__len__()}, allowed bounds : {self.bounds}"""))
            
    def _validate_items(self, value : typing.Dict[typing.Any, typing.Any]) -> None:
        keys = value.keys()
        values = value.values()
        if self.key_type is not None and len(keys) != 0:                    
            for key in keys:
                if not isinstance(key, self.key_type):
                    raise TypeError(wrap_error_text(f"""
                            Keys for typed dictionary contain incompatible types. 
                            Allowed types : {self.key_type}, given type : {type(key)}"""))
        if self.item_type is not None and len(values) != 0: 
            for value in values:
                if not isinstance(value, self.item_type):
                    raise TypeError(wrap_error_text(f"""
                            Values for typed dictionary contain incompatible types. 
                            Allowed types : {self.item_type}. given type : {type(value)}"""))
       
    def _validate_key_value_pair(self, __key : typing.Any, __value : typing.Any) -> None:
        if self.key_type is not None:
            if not isinstance(__key, self.key_type):
                raise TypeError("given key {} is not of {}.".format(__key, self.key_type))
        if self.item_type is not None: 
            if not isinstance(__value, self.item_type):
                raise TypeError("given item {} is not of {}.".format(__value, self.item_type))

    def __iter__(self) -> typing.Iterator:
        return self._inner.__iter__()

    def __setitem__(self, __key : typing.Any, __value : typing.Any) -> None:
        if self.constant:
            raise ValueError(f"Dict {get_iterable_printfriendly_repr(self._inner)} is a constant and cannot be modified.")
        if __key not in self._inner:
            self._validate_bounds_for_extension()
        self._validate_key_value_pair(__key, __value)
        self._inner.__setitem__(__key, __value)

    def __delitem__(self, __v : typing.Any) -> None:
        self._inner.__delitem__(__v)

    def __getitem__(self, __k : typing.Any) -> typing.Any:
        return self._inner.__getitem__(__k)

    def __str__(self) -> str:
        return self._inner.__str__()

    def __len__(self) -> int:
        return self._inner.__len__()

    def __contains__(self, __o : object) -> bool:
        return self._inner.__contains__(__o)

    def __eq__(self, __o: object) -> bool:
        return self._inner.__eq__(__o) 

    def __ne__(self, __o: object) -> bool:
        return self._inner.__ne__(__o)
    
    def __format__(self, __format_spec: str) -> str:
        return self._inner.__format__(__format_spec)        
    
    def __sizeof__(self) -> int:
        return self._inner.__sizeof__()
        
    def __repr__(self) -> str:
        return self._inner.__repr__()
        
    def fromkeys(self, __iterable, __value : typing.Any):
        return self._inner.fromkeys(__iterable, __value)

    def keys(self) -> typing.Any:
        return self._inner.keys()

    def items(self) -> typing.Any:
        return self._inner.items()

    def values(self) -> typing.Any:
        return self._inner.values()

    def get(self, __key : typing.Any, __default : typing.Any = None):
        return self._inner.get(__key, __default)

    def setdefault(self, __key : typing.Any) -> None:
        self._inner.setdefault(__key)       

    def clear(self) -> None:
        self._inner.clear()

    def copy(self, return_as_typed : bool = False) -> typing.Union["TypeConstrainedDict", typing.Dict]:
        if return_as_typed:
            return TypeConstrainedDict(default=self._inner.copy(), key_type=self.key_type, item_type=self.item_type, 
                bounds=self.bounds, constant=self.constant, skip_validate=True)
        else:
            return self._inner.copy()

    def popitem(self) -> tuple:
        return self._inner.popitem()
    
    def pop(self, __key : typing.Any) -> typing.Any:
        return self._inner.pop(__key)
    
    def update(self, __o : typing.Any) -> None:
        self._validate_for_insertion(__o)
        self._inner.update(__o)
        

class TypedKeyMappingsConstrainedDict(TypeConstrainedDict):

    def __init__(self, default: typing.Dict, *, type_mapping : typing.Dict, 
            allow_unspecified_keys : bool = False, bounds : tuple = (0, None), constant : bool = False, 
            skip_validate : bool = False) -> None:
        self.type_mapping = type_mapping
        self.allow_unspecified_keys = allow_unspecified_keys
        self.key_list = self.type_mapping.keys()
        super().__init__(default, key_type = None, item_type = None, bounds = bounds, constant = constant,
                skip_validate=skip_validate)

    def _validate_items(self, value: typing.Dict) -> None:
        for key, val in value.items():
            self._validate_key_value_pair(key, val)
                
    def _validate_key_value_pair(self, __key: typing.Any, __value: typing.Any) -> None:
        if __key not in self.key_list:
            if self.allow_unspecified_keys:
                pass 
            else: 
                raise KeyError(f"Keys except {self.key_list} not allowed for typed dictionary. Given key : {__key}.")
        elif not isinstance(__value, self.type_mapping[__key]):
            raise TypeError(wrap_error_text(f""" 
                Value for key {__key} not of expected type : {self.type_mapping[__key]}. Given type : {type(__value)}."""))
        
    def copy(self, return_as_typed : bool = False) -> typing.Union["TypedKeyMappingsConstrainedDict", typing.Dict]:
        if return_as_typed:
            return TypedKeyMappingsConstrainedDict(default=self._inner.copy(), type_mapping=self.type_mapping,  
                bounds=self.bounds, constant=self.constant, skip_validate=True)
        else:
            return self._inner.copy()
           

class TypedDict(ClassSelector):

    __slots__ = ['key_type', 'item_type', 'bounds']
    
    def __init__(self, default : typing.Optional[typing.Dict] = None, *, key_type : typing.Any = None, 
            item_type : typing.Any = None, deepcopy_default : bool = True, allow_None : bool = True,
            bounds : tuple = (0, None), doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, 
            per_instance_descriptor : bool = False, 
            class_member : bool = False, fget : typing.Optional[typing.Callable] = None, 
            fset : typing.Optional[typing.Callable] = None, fdel : typing.Optional[typing.Callable] = None,
            precedence : typing.Optional[float] = None) -> None:
        if default is not None:
            default = TypeConstrainedDict(default, key_type=key_type, item_type=item_type, bounds=bounds,
                        constant=constant, skip_validate=False) # type: ignore
        self.key_type = key_type
        self.item_type = item_type
        self.bounds = bounds 
        super().__init__(class_=TypeConstrainedDict, default=default, isinstance=True, deepcopy_default=deepcopy_default, 
            doc=doc, constant=constant, readonly=readonly, allow_None=allow_None, fget=fget, fset=fset, fdel=fdel, 
            per_instance_descriptor=per_instance_descriptor, class_member=class_member, precedence=precedence)

    def __set__(self, obj, value):
        if value is not None:
            container = TypeConstrainedDict(default=value, key_type=self.key_type, item_type=self.item_type,
                                bounds=self.bounds, constant=self.constant, skip_validate=False)     
            return super().__set__(obj, container) # re-set it to trigger param related activities
        else:
            return super().__set__(obj, value) # re-set it to trigger param related activities
     
    @classmethod
    def serialize(cls, value: TypeConstrainedDict) -> typing.Any:
        if value is None:
            return None
        return value._inner
  

class TypedKeyMappingsDict(ClassSelector):

    __slots__ = ['type_mapping', 'allow_unspecified_keys', 'bounds']
    
    def __init__(self, default : typing.Optional[typing.Dict[typing.Any, typing.Any]] = None, *, 
            type_mapping : typing.Dict, 
            allow_unspecified_keys : bool = True, bounds : tuple = (0, None), 
            deepcopy_default : bool = True, allow_None : bool = True,
            doc : typing.Optional[str] = None, constant : bool = False, readonly : bool = False, 
            per_instance_descriptor : bool = False, class_member : bool = False, 
            fget : typing.Optional[typing.Callable] = None, fset : typing.Optional[typing.Callable] = None,
            fdel : typing.Optional[typing.Callable] = None, precedence : typing.Optional[float] = None) -> None:
        if default is not None:
            default = TypedKeyMappingsConstrainedDict(default=default, type_mapping=type_mapping, 
                        allow_unspecified_keys=allow_unspecified_keys, bounds=bounds, constant=constant, 
                        skip_validate=False) # type: ignore 
        self.type_mapping = type_mapping
        self.allow_unspecified_keys = allow_unspecified_keys
        self.bounds = bounds 
        super().__init__(class_=TypedKeyMappingsConstrainedDict, default=default, 
            isinstance=True, deepcopy_default=deepcopy_default, doc=doc, constant=constant, readonly=readonly, 
            allow_None=allow_None, per_instance_descriptor=per_instance_descriptor, class_member=class_member, 
            fget=fget, fset=fset, fdel=fdel, precedence=precedence) 
                       
    def __set__(self, obj, value):
        if value is not None:
            container = TypedKeyMappingsConstrainedDict(default=value, type_mapping=self.type_mapping, 
                    allow_unspecified_keys=self.allow_unspecified_keys, bounds=self.bounds, constant=self.constant,
                    skip_validate=False)
            return super().__set__(obj, container) 
        else:
            return super().__set__(obj, value) # re-set it to trigger param related activities
        
    @classmethod
    def serialize(cls, value: TypeConstrainedDict) -> typing.Any:
        if value is None:
            return None
        return value._inner


def hashable(x):
    """
    Return a hashable version of the given object x, with lists and
    dictionaries converted to tuples.  Allows mutable objects to be
    used as a lookup key in cases where the object has not actually
    been mutated. Lookup will fail (appropriately) in cases where some
    part of the object has changed.  Does not (currently) recursively
    replace mutable subobjects.
    """
    if isinstance(x, collections.abc.MutableSequence):
        return tuple(x)
    elif isinstance(x, collections.abc.MutableMapping):
        return tuple([(k,v) for k,v in x.items()])
    else:
        return x


def named_objs(objlist, namesdict=None):
    """
    Given a list of objects, returns a dictionary mapping from
    string name for the object to the object itself. Accepts
    an optional name,obj dictionary, which will override any other
    name if that item is present in the dictionary.
    """
    objs = OrderedDict()

    objtoname = {}
    unhashables = []
    if namesdict is not None:
        for k, v in namesdict.items():
            try:
                objtoname[hashable(v)] = k
            except TypeError:
                unhashables.append((k, v))

    for obj in objlist:
        if objtoname and hashable(obj) in objtoname:
            k = objtoname[hashable(obj)]
        elif any(obj is v for (_, v) in unhashables):
            k = [k for (k, v) in unhashables if v is obj][0]
        elif hasattr(obj, "name"):
            k = obj.name
        elif hasattr(obj, '__name__'):
            k = obj.__name__
        else:
            k = str(obj)
        objs[k] = obj
    return objs


def _is_abstract(class_):
    try:
        return class_.__abstract
    except AttributeError:
        return False


# Could be a method of ClassSelector.
def concrete_descendents(parentclass):
    """
    Return a dictionary containing all subclasses of the specified
    parentclass, including the parentclass.  Only classes that are
    defined in scripts that have been run or modules that have been
    imported are included, so the caller will usually first do ``from
    package import *``.

    Only non-abstract classes will be included.
    """
    return dict((c.__name__, c) for c in descendents(parentclass)
                if not _is_abstract(c))



__all__ = [ 
            'String', 'Bytes', 'IPAddress', 'Number', 'Integer', 'Boolean', 'Iterable', 'Tuple', 'List', 'Callable', 
            'CSS3Color', 'Composite', 'Selector', 'ClassSelector', 'TupleSelector', 
            'Filename', 'Foldername', 'Path', 'FileSelector', 'MultiFileSelector', 
            'Date', 'CalendarDate',  'Range', 'DateRange', 'CalendarDateRange', 
            'TypeConstrainedList', 'TypeConstrainedDict', 'TypedKeyMappingsConstrainedDict', 
            'TypedList', 'TypedDict', 'TypedKeyMappingsDict', 
        ]



import typing
from .constants import JSON

class JSONSchemaError(Exception):
    """
    common error to be raised for JSON schema
    validation irrespective of internal validation used
    """
    pass 

class JSONValidationError(Exception):
    """
    common error to be raised for JSON validation
    irrespective of internal validation used
    """
    pass



class BaseSchemaValidator: # type definition
    """
    Base class for all schema validators. 
    Serves as a type definition. 
    """
    def __init__(self, schema : JSON) -> None:
        self.schema = schema

    def validate(self, data) -> None:
        """
        validate the data against the schema. 
        """
        raise NotImplementedError("validate method must be implemented by subclass")


try: 
    import fastjsonschema 
    
    class FastJsonSchemaValidator(BaseSchemaValidator):
        """
        JSON schema validator according to fast JSON schema.
        Useful for performance with dictionary based schema specification
        which msgspec has no built in support. Normally, for speed, 
        one should try to use msgspec's struct concept.
        """

        def __init__(self, schema : JSON) -> None:
            super().__init__(schema)
            self.validator = fastjsonschema.compile(schema)

        def validate(self, data) -> None:
            """validates and raises exception when failed directly to the caller"""
            try: 
                self.validator(data)
            except fastjsonschema.JsonSchemaException as ex: 
                raise JSONSchemaError(str(ex)) from None

        def json(self):
            """allows JSON (de-)serializable of the instance itself"""
            return self.schema

        def __get_state__(self):
            return self.schema
        
        def __set_state__(self, schema):
            return FastJsonSchemaValidator(schema)

except ImportError as ex:
    pass



import jsonschema

class JsonSchemaValidator(BaseSchemaValidator):
    """
    JSON schema validator according to standard python JSON schema.
    Somewhat slow, consider msgspec if possible. 
    """
    
    def __init__(self, schema) -> None:
        jsonschema.Draft7Validator.check_schema(schema)
        super().__init__(schema)
        self.validator = jsonschema.Draft7Validator(schema)

    def validate(self, data) -> None:
        self.validator.validate(data)

    def json(self):
        """allows JSON (de-)serializable of the instance itself"""
        return self.schema

    def __get_state__(self):
        return self.schema
    
    def __set_state__(self, schema):
        return JsonSchemaValidator(schema)
    


def _get_validator_from_user_options(option : typing.Optional[str] = None) -> BaseSchemaValidator:
    """
    returns a JSON schema validator based on user options
    """
    if option == "fastjsonschema":
        return FastJsonSchemaValidator
    elif option == "jsonschema" or not option:
        return JsonSchemaValidator
    else:
        raise ValueError(f"Unknown JSON schema validator option: {option}")


import logging
from contextlib import contextmanager
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL, Logger



VERBOSE = INFO - 1
logging.addLevelName(VERBOSE, "VERBOSE")
    
def get_logger(name : str = None) -> Logger:
    if name is None:
        root_logger = logging.getLogger('param')
        if not root_logger.handlers:
            root_logger.setLevel(logging.INFO)
            formatter = logging.Formatter(
                fmt='%(levelname)s:%(name)s: %(message)s')
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            root_logger.addHandler(handler)
            return root_logger
    else:
        return logging.getLogger('param.' + name)


@contextmanager
def logging_level(level : int):
    """
    Temporarily modify param's logging level.
    """
    level = level.upper()
    levels = [DEBUG, INFO, WARNING, ERROR, CRITICAL, VERBOSE]
    level_names = ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL', 'VERBOSE']

    if level not in level_names:
        raise Exception("Level %r not in %r" % (level, levels))

    param_logger = get_logger()
    logging_level = param_logger.getEffectiveLevel()
    param_logger.setLevel(levels[level_names.index(level)])
    try:
        yield None
    finally:
        param_logger.setLevel(logging_level)



__all__ = ['logging_level', 'get_logger']